import logging
from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestAudioDefaultSettings(MpfMcTestCase):
    """
    Tests the default audio settings (no 'sound_system' entries in the config file)
    """

    def get_machine_path(self):
        return 'tests/machine_files/audio'

    def get_config_file(self):
        return 'test_audio_default_settings.yaml'

    def test_default_sound_system(self):
        """ Tests the sound system and audio interface with default settings """

        if self.mc.sound_system is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system is not enabled - unable to run audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)

        if self.mc.sound_system.audio_interface is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(self.mc.sound_system.audio_interface)
        settings = self.mc.sound_system.audio_interface.get_settings()
        self.assertIsNotNone(settings)
        self.assertIn(settings['buffer_samples'], [1024, 2048])
        self.assertIn(settings['audio_channels'], [1, 2])
        self.assertEqual(settings['sample_rate'], 44100)

        # /sounds/sfx
        self.assertIn('198361_sfx-028', self.mc.sounds)     # .wav
        self.assertIn('210871_synthping', self.mc.sounds)   # .wav
        self.assertIn('264828_text', self.mc.sounds)        # .ogg
        self.assertIn('4832__zajo__drum07', self.mc.sounds)   # .wav
        self.assertIn('84480__zgump__drum-fx-4', self.mc.sounds)   # .wav
        self.assertIn('100184__menegass__rick-drum-bd-hard', self.mc.sounds)   # .wav

        # /sounds/voice
        self.assertIn('104457_moron_test', self.mc.sounds)  # .wav
        self.assertIn('113690_test', self.mc.sounds)        # .wav

        # Check for default ducking assigned only to sounds in the sfx folder
        self.assertTrue(self.mc.sounds['198361_sfx-028'].has_ducking)
        self.assertEqual(0, self.mc.sounds['198361_sfx-028'].ducking.delay)
        self.assertEqual(0.3, self.mc.sounds['198361_sfx-028'].ducking.attack)
        self.assertEqual(0.45, self.mc.sounds['198361_sfx-028'].ducking.attenuation)
        self.assertEqual(0.5, self.mc.sounds['198361_sfx-028'].ducking.release_point)
        self.assertEqual(1.0, self.mc.sounds['198361_sfx-028'].ducking.release)
        self.assertTrue(self.mc.sounds['210871_synthping'].has_ducking)
        self.assertTrue(self.mc.sounds['264828_text'].has_ducking)
        self.assertTrue(self.mc.sounds['4832__zajo__drum07'].has_ducking)
        self.assertTrue(self.mc.sounds['84480__zgump__drum-fx-4'].has_ducking)
        self.assertTrue(self.mc.sounds['100184__menegass__rick-drum-bd-hard'].has_ducking)

        # These sounds should not have ducking
        self.assertFalse(self.mc.sounds['104457_moron_test'].has_ducking)
        self.assertFalse(self.mc.sounds['113690_test'].has_ducking)



from functools import partial

from mpf.config_players.plugin_player import PluginPlayer
from mpf.core.utility_functions import Util
from mpfmc.config_collections.animation import AnimationCollection
from mpfmc.uix.widget_magic_events import magic_events


class MpfSlidePlayer(PluginPlayer):
    """Base class for part of the slide player which runs as part of MPF.

    Note: This class is loaded by MPF and everything in it is in the context of
    MPF, not the mpf-mc. MPF finds this instance because the mpf-mc setup.py
    has the following entry_point configured:

        slide_player=mpfmc.config_players.slide_player:register_with_mpf

    """
    config_file_section = 'slide_player'
    show_section = 'slides'

    def _validate_config_item(self, device, device_settings):
        # device is slide name, device_settings
        dict_is_widgets = False

        # if settings is list, it's widgets
        if isinstance(device_settings, list):
            device_settings = dict(widgets=device_settings)

        # Now check to see if all the settings are valid
        # slide settings. If not, assume it's a single widget settings.
        if isinstance(device_settings, dict):

            for key in device_settings.keys():
                if key not in self.machine.config_validator.get_config_spec()['slide_player']:
                    dict_is_widgets = True
                    break

            if dict_is_widgets:
                device_settings = dict(widgets=[device_settings])
        else:
            raise AssertionError("Settings in slide_player {} have to be dict.".format(device))

        # todo make transition manager validation static and use that here too

        device_settings = self.machine.config_validator.validate_config(
            'slide_player', device_settings)

        if 'transition' in device_settings:
            if not isinstance(device_settings['transition'], dict):
                device_settings['transition'] = dict(
                    type=device_settings['transition'])

            try:
                device_settings['transition'] = (
                    self.machine.config_validator.validate_config(
                        'transitions:{}'.format(device_settings['transition']['type']), device_settings['transition']))

            except KeyError:
                raise ValueError('transition: section of config requires a'
                                 ' "type:" setting')

        if 'widgets' in device_settings:
            device_settings['widgets'] = self.process_widgets(
                device_settings['widgets'])

        return_dict = dict()
        return_dict[device] = device_settings

        return return_dict

    def process_widgets(self, config):
        # config is localized to a specific widget section

        # This method is based on
        # mpfmc.core.config_collections.widget.Widget.process_config()

        # We can't use that one though because there are lots of other imports
        # that module does.

        # todo we could move that to a central location

        if isinstance(config, dict):
            config = [config]

        # Note that we don't reverse the order of the widgets here since
        # they'll be reversed when they're played

        widget_list = list()

        for widget in config:
            widget_list.append(self.process_widget(widget))

        return widget_list

    def _register_trigger(self, event, **kwargs):
        """Register trigger via BCP for MC."""
        del kwargs
        client = self.machine.bcp.transport.get_named_client("local_display")
        if client:
            self.machine.bcp.interface.add_registered_trigger_event_for_client(client, event)
        else:
            self.machine.events.add_handler("bcp_clients_connected", partial(self._register_trigger, event))

    def process_widget(self, config):
        # config is localized widget settings

        config['_default_settings'] = list(config.keys())

        try:
            self.machine.config_validator.validate_config('widgets:{}'.format(
                config['type']).lower(), config, base_spec='widgets:common')
        except KeyError:
            raise KeyError("Slide config validation error. Something is "
                           "wrong here: {}".format(config))

        if 'control_events' in config:
            for event_dict in config['control_events']:
                if event_dict["event"] not in magic_events:
                    self._register_trigger(event_dict["event"])

        if 'reset_animations_events' in config:
            for event_name in config['reset_animations_events']:
                if event_name not in magic_events:
                    self._register_trigger(event_name)

        if 'animations' in config:
            config['animations'] = self.process_animations(
                config['animations'])

        else:
            config['animations'] = None

        return config

    def process_animations(self, config):
        # config is localized to the slide's 'animations' section

        for event_name, event_settings in config.items():

            # make sure the event_name is registered as a trigger event so MPF
            # will send those events as triggers via BCP. But we don't want
            # to register magic events since those aren't real MPF events.
            if event_name not in magic_events:
                self._register_trigger(event_name)

            # str means it's a list of named animations
            if isinstance(event_settings, str):
                event_settings = Util.string_to_event_list(event_settings)

            # dict means it's a single set of settings for one animation step
            elif isinstance(event_settings, dict):
                event_settings = [event_settings]

            # ultimately we're producing a list of dicts, so build that list
            # as we iterate
            new_list = list()
            for settings in event_settings:
                new_list.append(AnimationCollection.process_animation(settings, self.machine.config_validator))

            config[event_name] = new_list

        return config


player_cls = MpfSlidePlayer


def register_with_mpf(machine):
    return 'slide', MpfSlidePlayer(machine)



#!python
#cython: embedsignature=True, language_level=3

from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
import cython
import logging
import time
from math import ceil
from heapq import heappush, heappop, heapify

from mpfmc.assets.sound import SoundInstance, SoundStealingMethod
from mpfmc.core.audio.sdl2 cimport *
from mpfmc.core.audio.gstreamer cimport *
from mpfmc.core.audio.track cimport *
from mpfmc.core.audio.inline cimport lerpU8, in_out_quad
from mpfmc.core.audio.notification_message cimport *
from mpfmc.core.audio.track_standard cimport *
from mpfmc.core.audio.audio_exception import AudioException


# ---------------------------------------------------------------------------
#    Settings
# ---------------------------------------------------------------------------

# The maximum number of consecutive null buffers to receive while streaming before
# terminating the sound (will cause drop outs)
DEF CONSECUTIVE_NULL_STREAMING_BUFFER_LIMIT = 2


# ---------------------------------------------------------------------------
#    TrackStandard class
# ---------------------------------------------------------------------------
cdef class TrackStandard(Track):
    """
    Track class
    """

    def __init__(self, object mc, object audio_callback_data, str name, int track_num, int buffer_size,
                 int max_simultaneous_sounds=MAX_SIMULTANEOUS_SOUNDS_DEFAULT,
                 float volume=1.0):
        """
        Constructor
        Args:
            mc: The media controller app
            audio_callback_data: The AudioCallbackData struct wrapped in a PyCapsule
            name: The track name
            track_num: The track number
            buffer_size: The length of the track audio buffer in bytes
            max_simultaneous_sounds: The maximum number of sounds that can be played simultaneously
                on the track
            volume: The track volume (0.0 to 1.0)
        """
        # IMPORTANT: Call super class init function to allocate track state memory!
        super().__init__(mc, audio_callback_data, name, track_num, buffer_size, volume)

        self.log = logging.getLogger("Track." + str(track_num) + ".TrackStandard." + name)

        SDL_LockAudio()

        # Dictionary of SoundInstance class objects keyed by SoundInstance.id
        self._playing_instances_by_id = dict()

        # Priority queue of SoundInstance objects waiting to be played
        self._sound_queue = list()

        # Set track type specific settings
        self.state.mix_callback_function = TrackStandard.mix_playing_sounds

        # Allocate memory for the specific track type state struct (TrackStandardState)
        self.type_state = <TrackStandardState*> PyMem_Malloc(sizeof(TrackStandardState))
        self.state.type_state = <void*>self.type_state

        # Make sure the number of simultaneous sounds is within the allowable range
        if max_simultaneous_sounds > MAX_SIMULTANEOUS_SOUNDS_LIMIT:
            self.log.warning("The maximum number of simultaneous sounds per track is %d",
                             MAX_SIMULTANEOUS_SOUNDS_LIMIT)
            max_simultaneous_sounds = MAX_SIMULTANEOUS_SOUNDS_LIMIT
        elif max_simultaneous_sounds < 1:
            self.log.warning("The minimum number of simultaneous sounds per track is 1")
            max_simultaneous_sounds = 1
        self._max_simultaneous_sounds = max_simultaneous_sounds
        self.type_state.sound_player_count = max_simultaneous_sounds

        # Allocate memory for the sound player structs needed for the desired number of
        # simultaneous sounds that can be played on the track.
        self.type_state.sound_players = <SoundPlayer*> PyMem_Malloc(self.type_state.sound_player_count * sizeof(SoundPlayer))

        # Initialize sound player attributes
        for i in range(self.type_state.sound_player_count):
            self.type_state.sound_players[i].status = player_idle
            self.type_state.sound_players[i].track_num = self.number
            self.type_state.sound_players[i].number = i
            self.type_state.sound_players[i].current.sample = NULL
            self.type_state.sound_players[i].current.loops_remaining = 0
            self.type_state.sound_players[i].current.current_loop = 0
            self.type_state.sound_players[i].current.volume = 0
            self.type_state.sound_players[i].current.sample_pos = 0
            self.type_state.sound_players[i].current.loop_start_pos = 0
            self.type_state.sound_players[i].current.loop_end_pos = 0
            self.type_state.sound_players[i].current.sound_id = 0
            self.type_state.sound_players[i].current.sound_instance_id = 0
            self.type_state.sound_players[i].current.sound_priority = 0
            self.type_state.sound_players[i].current.fading_status = fading_status_not_fading
            self.type_state.sound_players[i].current.about_to_finish_marker = no_marker
            self.type_state.sound_players[i].current.sound_has_ducking = False
            self.type_state.sound_players[i].current.ducking_stage = ducking_stage_idle
            self.type_state.sound_players[i].current.ducking_control_points = g_array_sized_new(False, False, sizeof(guint8), CONTROL_POINTS_PER_BUFFER)
            self.type_state.sound_players[i].current.marker_count = 0
            self.type_state.sound_players[i].current.markers = g_array_new(False, False, sizeof(guint))
            self.type_state.sound_players[i].next.sample = NULL
            self.type_state.sound_players[i].next.loops_remaining = 0
            self.type_state.sound_players[i].next.current_loop = 0
            self.type_state.sound_players[i].next.volume = 0
            self.type_state.sound_players[i].next.sample_pos = 0
            self.type_state.sound_players[i].next.loop_start_pos = 0
            self.type_state.sound_players[i].next.loop_end_pos = 0
            self.type_state.sound_players[i].next.sound_id = 0
            self.type_state.sound_players[i].next.sound_instance_id = 0
            self.type_state.sound_players[i].next.sound_priority = 0
            self.type_state.sound_players[i].next.fading_status = fading_status_not_fading
            self.type_state.sound_players[i].next.about_to_finish_marker = no_marker
            self.type_state.sound_players[i].next.sound_has_ducking = False
            self.type_state.sound_players[i].next.ducking_stage = ducking_stage_idle
            self.type_state.sound_players[i].next.ducking_control_points = g_array_sized_new(False, False, sizeof(guint8), CONTROL_POINTS_PER_BUFFER)
            self.type_state.sound_players[i].next.marker_count = 0
            self.type_state.sound_players[i].next.markers = g_array_new(False, False, sizeof(guint))

        self.log.debug("Created Track %d %s with the following settings: "
                       "simultaneous_sounds = %d, volume = %f",
                       self.number, self.name, self.max_simultaneous_sounds, self.volume)

        SDL_UnlockAudio()

    def __dealloc__(self):
        """Destructor"""

        SDL_LockAudio()

        # Free the specific track type state and other allocated memory
        if self.type_state != NULL:
            for i in range(self.type_state.sound_player_count):
                g_array_free(self.type_state.sound_players[i].current.ducking_control_points, True)
                g_array_free(self.type_state.sound_players[i].next.ducking_control_points, True)
                g_array_free(self.type_state.sound_players[i].current.markers, True)
                g_array_free(self.type_state.sound_players[i].next.markers, True)

            PyMem_Free(self.type_state.sound_players)
            PyMem_Free(self.type_state)
            self.type_state = NULL
            if self.state != NULL:
                self.state.type_state = NULL

        SDL_UnlockAudio()

    def __repr__(self):
        return '<Track.{}.Standard.{}>'.format(self.number, self.name)

    @property
    def type(self):
        return "standard"

    @property
    def supports_in_memory_sounds(self):
        """Return whether or not track accepts in-memory sounds"""
        return True

    @property
    def supports_streaming_sounds(self):
        """Return whether or not track accepts streaming sounds"""
        return True

    @property
    def max_simultaneous_sounds(self):
        """Return the number of sounds that can be played simultaneously on this track"""
        return self._max_simultaneous_sounds

    cdef int _get_idle_sound_player(self):
        """
        Returns the index of the first idle sound player on the track.  If all
        players are currently busy playing, -1 is returned.
        """
        SDL_LockAudio()

        for index in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[index].status == player_idle:
                SDL_UnlockAudio()
                return index

        SDL_UnlockAudio()
        return -1

    def process(self):
        """Processes the track queue each tick."""

        cdef bint keep_checking = True
        cdef int idle_sound_player
        cdef GSList *iterator = NULL

        # Lock the mutex to ensure no audio data is changed during the playback processing
        # (multi-threaded protection)
        SDL_LockAudio()

        while keep_checking:
            # See if there are now any idle sound players
            idle_sound_player = self._get_idle_sound_player()
            if idle_sound_player >= 0:
                # Found an idle player, check if there are any sounds queued for playback
                sound_instance = self._get_next_sound()

                if sound_instance is not None:
                    self.log.debug("Getting sound from queue %s", sound_instance)
                    self._play_sound_on_sound_player(sound_instance=sound_instance, player=idle_sound_player)
                else:
                    keep_checking = False
            else:
                keep_checking = False

        # Process track notification messages
        if self.state.notification_messages != NULL:
            self.state.notification_messages = g_slist_reverse(self.state.notification_messages)
            iterator = self.state.notification_messages
            while iterator != NULL:
                self.process_notification_message(<NotificationMessageContainer*>iterator.data)
                g_slice_free1(sizeof(NotificationMessageContainer), iterator.data)
                iterator = iterator.next

            g_slist_free(self.state.notification_messages)
            self.state.notification_messages = NULL

        # Unlock the mutex since we are done accessing the audio data
        SDL_UnlockAudio()

    cdef process_notification_message(self, NotificationMessageContainer *notification_message):
        """Process a notification message to this track"""

        if notification_message == NULL:
            return

        SDL_LockAudio()

        # Check for track notification messages first (they do not need sound instance information)
        if notification_message.message in (notification_track_stopped, notification_track_paused):
            if notification_message.message == notification_track_stopped:
                self._reset_state()
                # Trigger any events
                if self.events_when_stopped is not None:
                    for event in self.events_when_stopped:
                        self.mc.post_mc_native_event(event, track=self._name)

            elif notification_message.message == notification_track_paused:
                # Trigger any events
                if self.events_when_paused is not None:
                    for event in self.events_when_paused:
                        self.mc.post_mc_native_event(event, track=self._name)
                pass

            SDL_UnlockAudio()
            return

        self.log.debug("Processing notification message %d for sound instance (id: %d)",
                       notification_message.message, notification_message.sound_instance_id)

        if notification_message.sound_instance_id not in self._playing_instances_by_id:
            self.log.warning("Received a notification message for a sound instance (id: %d) "
                             "that is no longer managed in the audio library. "
                             "Notification will be discarded.",
                             notification_message.sound_instance_id)

        elif notification_message.message == notification_sound_started:
            sound_instance = self._playing_instances_by_id[notification_message.sound_instance_id]
            if sound_instance is not None:
                sound_instance.set_playing()

        elif notification_message.message == notification_sound_stopped:
            sound_instance = self._playing_instances_by_id[notification_message.sound_instance_id]
            if sound_instance is not None:
                sound_instance.set_stopped()
                self.log.debug("Removing sound instance %s from playing sound "
                               "instance dictionary", str(sound_instance))
                del self._playing_instances_by_id[sound_instance.id]

        elif notification_message.message == notification_sound_looping:
            sound_instance = self._playing_instances_by_id[notification_message.sound_instance_id]
            if sound_instance is not None:
                sound_instance.set_looping()

        elif notification_message.message == notification_sound_about_to_finish:
            sound_instance = self._playing_instances_by_id[notification_message.sound_instance_id]
            if sound_instance is not None:
                sound_instance.set_about_to_finish()

        elif notification_message.message == notification_sound_marker:
            sound_instance = self._playing_instances_by_id[notification_message.sound_instance_id]
            if sound_instance is not None:
                sound_instance.set_marker(notification_message.data.marker.id)
        else:
            raise AudioException("Unknown notification message received on %s track", self.name)

        SDL_UnlockAudio()

    def _get_next_sound(self):
        """
        Returns the next sound in the priority queue ready for playback.

        Returns: A SoundInstance object. If the queue is empty, None is returned.

        This method ensures that the sound that is returned has not expired.
        If the next sound in the queue has expired, it is discarded and the
        next sound that has not expired is returned.
        """
        # We don't want to go through the entire sound queue more than once
        # in this method so keep track of the entry ids of the items we've
        # processed.  Once an item has been processed and retrieved again,
        # we are done and return None.
        cdef list sound_instances_retrieved_from_queue = list()
        while True:
            # Return none if sound queue is empty
            if len(self._sound_queue) == 0:
                return None

            # Get the next item in the queue (sorted by priority and expiration time)
            sound_instance = heappop(self._sound_queue)

            # Check if we've already processed the sound instance during this call (if
            # so, put it back in the queue and return)
            if sound_instance in sound_instances_retrieved_from_queue:
                heappush(self._sound_queue, sound_instance)
                return None

            # Keep track of entries we've processed during this call
            sound_instances_retrieved_from_queue.append(sound_instance)

            # If the sound is still loading and not expired, put it back in the queue
            if not sound_instance.sound.loaded and sound_instance.sound.loading and \
                    (sound_instance.exp_time is None or sound_instance.exp_time > time.time()):
                heappush(self._sound_queue, sound_instance)
                self.log.debug("Next pending sound in queue is still loading, "
                               "re-queueing sound %s",
                               sound_instance)
            else:
                # Return the next sound from the priority queue if it has not expired
                if sound_instance.exp_time is None or sound_instance.exp_time > time.time():
                    self.log.debug("Retrieving next pending sound from queue %s", sound_instance)
                    sound_instance.set_pending()  # Notify sound instance it is no longer queued
                    return sound_instance
                else:
                    self.log.debug("Discarding expired sound from queue %s", sound_instance)
                    sound_instance.set_expired()  # Notify sound instance it has expired

    def _remove_sound_from_queue(self, sound not None):
        """
        Removes all sound instances of the specify sound from the priority sound queue.

        Args:
            sound: The sound object to remove
        """
        cdef list sound_queue_copy = self._sound_queue.copy()
        for sound_instance in sound_queue_copy:
            if sound_instance.sound_id == sound.id:
                self._sound_queue.remove(sound_instance)
                sound_instance.set_canceled()

        heapify(self._sound_queue)

    def _remove_sound_instance_from_queue(self, sound_instance not None):
        """
        Removes a sound from the priority sound queue.
        Args:
            sound_instance: The sound instance object to remove
        """
        try:
            self._sound_queue.remove(sound_instance)
            self.log.debug("Removing pending sound from queue %s", sound_instance)
            sound_instance.set_canceled()
            heapify(self._sound_queue)
        except ValueError:
            pass

    def _remove_all_sounds_with_context_from_queue(self, context):
        """Removes all sounds with the specified context from the priority sound queue.
        """
        for sound_instance in self._sound_queue.copy():
            if sound_instance.context == context:
                self.log.debug("Removing pending sound with context %s from queue %s", context, sound_instance)
                sound_instance.set_canceled()
                self._sound_queue.remove(sound_instance)

        heapify(self._sound_queue)

    def _remove_all_sounds_with_key_from_queue(self, key):
        """Removes all sounds with the specified key from the priority sound queue.
        """
        for sound_instance in self._sound_queue.copy():
            if sound_instance.key == key:
                self.log.debug("Removing pending sound with key %s from queue %s", key, sound_instance)
                sound_instance.set_canceled()
                self._sound_queue.remove(sound_instance)

        heapify(self._sound_queue)

    def _remove_all_sounds_from_queue(self):
        """Removes all sounds from the priority sound queue.
        """
        for sound_instance in self._sound_queue:
            self.log.debug("Removing pending sound from queue %s", sound_instance)
            sound_instance.set_canceled()

        self._sound_queue.clear()

    cdef int _get_playing_sound_count(self, Uint64 sound_id):
        """Return the number of currently playing instances of the given sound id"""
        cdef int count = 0
        SDL_LockAudio()

        for i in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[i].status != player_idle and \
                            self.type_state.sound_players[i].current.sound_id == sound_id:
                count += 1

        SDL_UnlockAudio()
        return count

    cdef list _get_playing_sound_instances(self, Uint64 sound_id):
        """Return the list of currently playing instances of the given sound id"""
        cdef list instances = list()
        cdef Uint64 instance_id

        SDL_LockAudio()

        for i in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[i].status != player_idle and \
                            self.type_state.sound_players[i].current.sound_id == sound_id:

                if self.type_state.sound_players[i].status == player_replacing:
                    instance_id = self.type_state.sound_players[i].next.sound_instance_id
                else:
                    instance_id = self.type_state.sound_players[i].current.sound_instance_id

                if instance_id in self._playing_instances_by_id:
                    instances.append(self._playing_instances_by_id[instance_id])

        SDL_UnlockAudio()
        return instances

    def _get_oldest_playing_sound_instance(self, Uint64 sound_id):
        """Return the oldest sound instance currently playing"""
        cdef list playing_instances = self._get_playing_sound_instances(sound_id)

        if not playing_instances:
            return None

        oldest_instance = playing_instances[0]
        for sound_instance in playing_instances[1:]:
            if sound_instance.timestamp < oldest_instance.timestamp:
                oldest_instance = sound_instance

        return oldest_instance

    def _get_newest_playing_sound_instance(self, Uint64 sound_id):
        """Return the newest sound instance currently playing"""
        cdef list playing_instances = self._get_playing_sound_instances(sound_id)

        if not playing_instances:
            return None

        newest_instance = playing_instances[0]
        for sound_instance in playing_instances[1:]:
            if sound_instance.timestamp > newest_instance.timestamp:
                newest_instance = sound_instance

        return newest_instance

    def play_sound(self, sound not None, context=None, settings=None):
        """
        Plays a sound on the current track.
        Args:
            sound: The SoundAsset or SoundPool object to play
            context: The context from which the sound is played.
            settings: A dictionary of settings for playback
        Returns:
            A SoundInstance object if the sound will be played (or queued for playback).
            None if the sound could not be played.
        """
        self.log.debug("play_sound - Processing sound '%s' for playback.", sound.name)

        SDL_LockAudio()

        # A try/else/finally block is used here to ensure SDL_UnlockAudio is always called
        # before the function returns (since there are so many return branches it is easier
        # to use than placing the SDL_UnlockAudio call before every return).
        try:

            # Sound cannot be played if the track is stopped or paused
            if self.state.status == track_status_stopped or self.state.status == track_status_paused:
                self.log.debug("play_sound - %s track is not currently playing and "
                               "therefore the request to play sound %s will be canceled",
                               self.name, sound.name)
                return None

            sound_instance = None

            # Check if sound can be played based on if the maximum number of playing instances of
            # this sound has been reached or not.
            if sound.simultaneous_limit is not None and self._get_playing_sound_count(sound.id) >= sound.simultaneous_limit:

                # Maximum number of simultaneous instances of the specified sound has been reached.
                # Perform action based on sound stealing method
                if sound.stealing_method == SoundStealingMethod.oldest:
                    oldest_instance = self._get_oldest_playing_sound_instance(sound.id)
                    if oldest_instance is not None:
                        sound_instance = SoundInstance(sound, context, settings)
                        self._replace_sound_instance(oldest_instance, sound_instance)
                        self.log.debug("Sound %s has reached the maximum number of instances. "
                                       "Replacing oldest instance", sound.name)
                        return sound_instance

                elif sound.stealing_method == SoundStealingMethod.newest:
                    newest_instance = self._get_newest_playing_sound_instance(sound.id)
                    if newest_instance is not None:
                        sound_instance = SoundInstance(sound, context, settings)
                        self._replace_sound_instance(newest_instance, sound_instance)
                        self.log.debug("Sound %s has reached the maximum number of instances. "
                                       "Replacing newest instance", sound.name)
                        return sound_instance

                else:
                    # New instance will not be played; it will be skipped
                    self.log.debug("Sound %s has reached the maximum number of instances. "
                                   "Sound will be skipped", sound.name)
                    return None
            else:
                sound_instance = SoundInstance(sound, context, settings)

            if sound_instance.max_queue_time is None:
                sound_instance.exp_time = None
            else:
                sound_instance.exp_time = time.time() + sound_instance.max_queue_time

            # Make sure sound is loaded.  If not, we assume the sound is being loaded and we
            # add it to the queue so it will be picked up on the next loop.
            if not sound_instance.sound.loaded:
                # If the sound is not already loading, load it now
                if not sound_instance.sound.loading:
                    sound_instance.sound.load()

                if sound_instance.max_queue_time != 0:
                    self._queue_sound(sound_instance)
                    self.log.debug("play_sound - Sound %s was not loaded and therefore has been "
                                   "queued for playback.", sound_instance.name)
                else:
                    self.log.debug("play_sound - Sound %s was not loaded and max_queue_time = 0, "
                                   "therefore it has been discarded and will not be played.", sound_instance.name)
                    sound_instance.set_expired()
                    return None

            else:
                # The sound is loaded and ready for playback.
                # If the sound can be played right away (available player) then play it.
                # Is there an available sound player?
                sound_player = self._get_sound_player_with_lowest_priority()
                player = sound_player[0]
                lowest_priority = sound_player[1]

                if lowest_priority is None:
                    self.log.debug("play_sound - Sound player %d is available "
                                   "for playback", player)
                    # Play the sound using the available player
                    self._play_sound_on_sound_player(sound_instance=sound_instance, player=player)
                else:
                    # All sound players are currently busy:
                    self.log.debug("play_sound - No idle sound player is available.")
                    self.log.debug("play_sound - Sound player %d is currently playing the sound with "
                                   "the lowest priority (%d).", player, lowest_priority)

                    # If the lowest priority of all the sounds currently playing is lower than
                    # the requested sound, kill the lowest priority sound and replace it.
                    if sound_instance.priority > lowest_priority:
                        self.log.debug("play_sound - Sound priority (%d) is higher than the "
                                       "lowest sound currently playing (%d). Forcing playback "
                                       "on sound player %d.", sound_instance.priority, lowest_priority, player)
                        self._play_sound_on_sound_player(sound_instance=sound_instance,
                                                         player=player,
                                                         force=True)

                    elif sound_instance.max_queue_time == 0:
                        # The sound could not be played immediately and has now expired (max_queue_time == 0)
                        self.log.debug("play_sound - Sound priority (%d) is less than or equal to the "
                                       "lowest sound currently playing (%d). Sound could not be played"
                                       "immediately and has now expired (max_queue_time = 0) and will "
                                       "not be played.",
                                       sound_instance.priority, lowest_priority)
                        sound_instance.set_expired()
                        return None

                    else:
                        # Add the requested sound to the priority queue
                        self.log.debug("play_sound - Sound priority (%d) is less than or equal to the "
                                       "lowest sound currently playing (%d). Sound will be queued "
                                       "for playback.", sound_instance.priority, lowest_priority)
                        self._queue_sound(sound_instance)

        except Exception as ex:
            # An exception occurred, sound could not be played
            self.log.error("Track %s: play_sound encountered an unexpected exception while "
                           "attempting to play the %s sound: %s", self.name, sound.name, ex)
            raise AudioException("Track {} play_sound encountered an unexpected exception while "
                                 "attempting to play the {} sound.".format(self.name, sound.name)) from ex

        else:
            # No exception occurred, return the newly created sound instance that will be played
            return sound_instance

        finally:
            # Always unlock the audio mutex before returning
            SDL_UnlockAudio()

    def _replace_sound_instance(self, old_instance not None, sound_instance not None):
        """
        Replace a currently playing instance with another sound instance.
        Args:
            old_instance: The currently playing sound instance to replace
            sound_instance: The new sound instance to begin playing immediately
        """

        self.log.debug("replace_sound_instance - Preparing to replace existing sound with a new sound instance")

        # Find which player is currently playing the specified sound instance to replace
        SDL_LockAudio()
        player = self._get_player_playing_sound_instance(old_instance)

        if player >= 0:
            self._play_sound_on_sound_player(sound_instance, player, force=True)
        else:
            self.log.debug("replace_sound_instance - Could not locate specified sound instance to replace")
            sound_instance.set_canceled()

        SDL_UnlockAudio()

    def _queue_sound(self, sound_instance not None):
        """Adds a sound to the queue to be played when a sound player becomes available.

        Args:
            sound_instance: The SoundInstance object to play.

        Note that this method will insert this sound into a position in the
        queue based on its priority, so highest-priority sounds are played
        first.
        """
        heappush(self._sound_queue, sound_instance)

        # Notify sound instance it has been queued
        sound_instance.set_queued()
        self.log.debug("Queueing sound %s", sound_instance)

    def _get_sound_instances_for_sound(self, sound not None):
        """Return list of sound instances of the given sound."""

        cdef list instances = list()

        SDL_LockAudio()

        for i in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[i].status != player_idle and self.type_state.sound_players[
                i].current.sound_id == sound.id:

                sound_instance_id = self.type_state.sound_players[i].current.sound_instance_id
                if sound_instance_id in self._playing_instances_by_id:
                    instances.append(self._playing_instances_by_id[sound_instance_id])

        SDL_UnlockAudio()

        return instances

    def stop_sound(self, sound not None, fade_out=None):
        """
        Stops all instances of the specified sound immediately on the track. Any queued instances
        will be removed from the queue.
        Args:
            sound: The Sound to stop
            fade_out: Optional amount of time (seconds) to fade out before stopping
        """
        for sound_instance in self._get_sound_instances_for_sound(sound):
            self.stop_sound_instance(sound_instance, fade_out)

        self._remove_sound_from_queue(sound)

    def stop_sound_instance(self, sound_instance not None, fade_out=None):
        """
        Stops the specified sound instance immediately on the track. Any queued instances
        will be removed from the queue.
        Args:
            sound_instance: The SoundInstance to stop
            fade_out: Optional amount of time (seconds) to fade out before stopping
        """
        cdef SoundPlayer *player

        SDL_LockAudio()

        self.log.debug("Stopping sound %s and removing any pending instances from queue", sound_instance.name)

        if fade_out is None:
            fade_out = sound_instance.fade_out

        for i in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[i].status != player_idle and self.type_state.sound_players[
                i].current.sound_instance_id == sound_instance.id:

                # Update player to stop playing sound
                player = cython.address(self.type_state.sound_players[i])

                # Calculate fade out (if necessary)
                player.current.fade_steps_remaining = fade_out * self.state.callback_data.seconds_to_bytes_factor // self.state.callback_data.bytes_per_control_point
                if player.current.fade_steps_remaining > 0:
                    player.current.fade_out_steps = player.current.fade_steps_remaining
                    player.current.fading_status = fading_status_fading_out
                    player.status = player_stopping
                else:
                    # Sound will stop immediately - send sound stopped notification
                    send_sound_stopped_notification(player.number, player.current.sound_id, player.current.sound_instance_id, self.state)
                    player.status = player_idle

                # Adjust ducking release (if necessary)
                if player.current.sound_has_ducking:
                    # player.current.ducking_settings.release_duration = min(sound_instance.ducking.release * self.state.callback_data.seconds_to_bytes_factor, request_message.data.stop.fade_out_duration)
                    # player.current.ducking_settings.release_start_pos = player.current.sample_pos
                    # TODO: Add more intelligent ducking release point calculation here:
                    #       Take into consideration whether ducking is already in progress and when it was
                    #       originally scheduled to finish.
                    pass

        # Remove any instances of the specified sound that are pending in the sound queue
        self._remove_sound_instance_from_queue(sound_instance)

        SDL_UnlockAudio()

    def stop_sound_looping(self, sound not None):
        """
        Cancels looping for all instances of the specified sound on the track. Any queued instances
        will be removed from the queue.
        Args:
            sound: The Sound to stop looping
        """
        for sound_instance in self._get_sound_instances_for_sound(sound):
            self.stop_sound_instance_looping(sound_instance)

        self._remove_sound_from_queue(sound)

    def stop_sound_instance_looping(self, sound_instance not None):
        """
        Stops the specified sound instance on the track after the current loop iteration
        has finished. Any queued instances of the sound will be removed.
        Args:
            sound_instance: The Sound to stop
        """

        SDL_LockAudio()

        for i in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[i].status != player_idle and self.type_state.sound_players[
                i].current.sound_instance_id == sound_instance.id:
                # Set sound's loops_remaining variable to zero and loop end position to the end of the sound
                self.type_state.sound_players[i].current.loops_remaining = 0
                self.type_state.sound_players[i].current.loop_end_pos = self.type_state.sound_players[
                    i].current.sample.data.memory.size

        # Remove any instances of the specified sound that are pending in the sound queue.
        if self.sound_instance_is_in_queue(sound_instance):
            self._remove_sound_instance_from_queue(sound_instance)

        SDL_UnlockAudio()

    def clear_context(self, context):
        """
        Triggers the end mode action for current sound instances played from the specified context.
        Any queued instances with a matching context will be canceled (removed).

        Args:
            context: The context to clear
        """
        # If this track is managed by a playlist controller, do not clear the context as that
        # will be handled by the playlist controller.
        if self.mc.sound_system.audio_interface.get_playlist_controller(self._name):
            self.log.debug("Skip clearing context %s (playlist controller will handle it)", context)
            return

        self.log.debug("Clearing context %s", context)

        SDL_LockAudio()

        # Clear any instances in the queue with the given context value
        self._remove_all_sounds_with_context_from_queue(context)

        for sound_instance in list(self._playing_instances_by_id.values()):
            if sound_instance.context == context:
                # Stop or stop looping the sound instance (depends upon mode end action value)
                if sound_instance.stop_on_mode_end:
                    self.stop_sound_instance(sound_instance)
                else:
                    self.stop_sound_instance_looping(sound_instance)

        SDL_UnlockAudio()

    def _reset_state(self):
        """Resets the track state (stops all sounds immediately and clears the queue)"""
        SDL_LockAudio()

        self.log.debug("Resetting track state (sounds will be stopped and queue cleared")

        for i in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[i].status != player_idle:
                # Set stop sound event
                send_sound_stopped_notification(i,
                                                self.type_state.sound_players[i].current.sound_id,
                                                self.type_state.sound_players[i].current.sound_instance_id,
                                                self.state)
                self.type_state.sound_players[i].status = player_idle

        # Remove all sounds that are pending in the sound queue.
        self._remove_all_sounds_from_queue()

        SDL_UnlockAudio()

    def stop_all_sounds(self, float fade_out_seconds = 0.0):
        """
        Stops all playing sounds immediately on the track.
        Args:
            fade_out_seconds: The number of seconds to fade out the sounds before stopping
        """
        SDL_LockAudio()

        self.log.debug("Stopping all sounds and removing any pending sounds from queue")

        for i in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[i].status != player_idle:

                # Update player to stop playing sound
                player = cython.address(self.type_state.sound_players[i])

                # Calculate fade out (if necessary)
                player.current.fade_steps_remaining = <Uint32>(fade_out_seconds * self.state.callback_data.seconds_to_bytes_factor // self.state.callback_data.bytes_per_control_point)
                if player.current.fade_steps_remaining > 0:
                    player.current.fade_out_steps = player.current.fade_steps_remaining
                    player.current.fading_status = fading_status_fading_out
                    player.status = player_stopping
                else:
                    # Sound will stop immediately - send sound stopped notification
                    send_sound_stopped_notification(player.number, player.current.sound_id, player.current.sound_instance_id, self.state)
                    player.status = player_idle

                # Adjust ducking release (if necessary)
                if player.current.sound_has_ducking:
                    # player.current.ducking_settings.release_duration = min(sound_instance.ducking.release * self.state.callback_data.seconds_to_bytes_factor, request_message.data.stop.fade_out_duration)
                    # player.current.ducking_settings.release_start_pos = player.current.sample_pos
                    # TODO: Add more intelligent ducking release point calculation here:
                    #       Take into consideration whether ducking is already in progress and when it was
                    #       originally scheduled to finish.
                    pass

        # Remove all sounds that are pending in the sound queue.
        self._remove_all_sounds_from_queue()

        SDL_UnlockAudio()

    cdef tuple _get_sound_player_with_lowest_priority(self):
        """
        Retrieves the sound player currently with the lowest priority.

        Returns:
            A tuple consisting of the sound player index and the priority of
            the sound playing on that player (or None if the player is idle).

        """
        SDL_LockAudio()

        cdef int lowest_priority = 2147483647
        cdef int sound_player = -1
        cdef int i

        for i in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[i].status == player_idle:
                SDL_UnlockAudio()
                return i, None
            elif self.type_state.sound_players[i].current.sound_priority < lowest_priority:
                lowest_priority = self.type_state.sound_players[i].current.sound_priority
                sound_player = i

        SDL_UnlockAudio()
        return sound_player, lowest_priority

    cdef bint _play_sound_on_sound_player(self, sound_instance, int player, bint force=False):
        """
        Plays a sound using the specified sound player

        Args:
            sound_instance: The SoundInstance object to play
            player: The player number to use to play the sound
            force: Flag indicating whether or not the sound should be forced to play if
                the player is already busy playing another sound.

        Returns:
            True if sound instance was able to be played, False otherwise
        """
        self.log.debug("_play_sound_on_sound_player: %s, %s, %s", str(sound_instance), str(player), str(force))

        SDL_LockAudio()

        # The sound cannot be played if the track is stopped or paused
        if self.state.status == track_status_stopped or self.state.status == track_status_paused:
            self.log.debug("_play_sound_on_sound_player - %s track is not currently playing and "
                           "therefore the request to play sound %s will be canceled",
                           self.name, sound_instance.name)
            sound_instance.set_canceled()
            SDL_UnlockAudio()
            return False

        if not sound_instance.sound.loaded:
            self.log.debug("Specified sound is not loaded, could not "
                           "play sound %s", sound_instance.name)
            SDL_UnlockAudio()
            return False

        # Make sure the player in range
        if player in range(self.type_state.sound_player_count):

            # If the specified sound player is not idle do not play the sound if force is not set
            if self.type_state.sound_players[player].status != player_idle and not force:
                self.log.debug("All sound players are currently in use, "
                               "could not play sound %s", sound_instance.name)
                SDL_UnlockAudio()
                return False

            # Add sound to the dictionary of active sound instances
            self.log.debug("Adding sound instance %s to active sound dictionary", str(sound_instance))
            self._playing_instances_by_id[sound_instance.id] = sound_instance

            # Check if sound player is idle
            if self.type_state.sound_players[player].status == player_idle:
                # Start the player playing the sound instance
                self._set_player_playing(cython.address(self.type_state.sound_players[player]), sound_instance)
            else:
                # The player is currently busy playing another sound, force it to be replaced with the sound instance
                self._set_player_replacing(cython.address(self.type_state.sound_players[player]), sound_instance)

            self.log.debug("Sound %s is set to begin playback on standard track (loops=%d)",
                           sound_instance.name, sound_instance.loops)

            SDL_UnlockAudio()
            return True

        SDL_UnlockAudio()
        return False

    cdef _set_player_sound_settings(self, SoundSettings *sound_settings, object sound_instance):
        """
        Set sound settings for the player from the sound instance.
        Args:
            sound_settings: A pointer to a SoundSettings structure
            sound_instance: The sound instance
        """
        cdef int bytes_per_sample_frame

        if sound_settings == NULL or sound_instance is None:
            return

        # Get the sound sample buffer container
        cdef SoundFile sound_container = sound_instance.container

        # Setup the player to start playing the sound
        sound_settings.sample_pos = <Uint32>(sound_instance.start_at * self.state.callback_data.seconds_to_bytes_factor)
        sound_settings.sample_pos = self._fix_sample_frame_pos(sound_settings.sample_pos, self.state.callback_data.bytes_per_sample, self.state.callback_data.channels)
        if sound_settings.sample_pos > sound_container.sample.size:
            sound_settings.sample_pos = 0

        # Setup loop start position (defaults to 0)
        sound_settings.loop_start_pos = <Uint32>(sound_instance.loop_start_at * self.state.callback_data.seconds_to_bytes_factor)
        sound_settings.loop_start_pos = self._fix_sample_frame_pos(sound_settings.loop_start_pos, self.state.callback_data.bytes_per_sample, self.state.callback_data.channels)
        if sound_settings.loop_start_pos > sound_container.sample.size:
            sound_settings.loop_start_pos = 0

        # Setup loop end position (if loop_end_at is None, set loop end to end of sound)
        if sound_instance.loop_end_at is None:
            sound_settings.loop_end_pos = sound_container.sample.size
        else:
            sound_settings.loop_end_pos = <Uint32>(sound_instance.loop_end_at * self.state.callback_data.seconds_to_bytes_factor)
            sound_settings.loop_end_pos = self._fix_sample_frame_pos(sound_settings.loop_end_pos, self.state.callback_data.bytes_per_sample, self.state.callback_data.channels)
            if sound_settings.loop_end_pos > sound_container.sample.size:
                sound_settings.loop_end_pos = sound_container.sample.size

        sound_settings.current_loop = 0
        sound_settings.sound_id = sound_instance.sound_id
        sound_settings.sound_instance_id = sound_instance.id
        sound_settings.sample = cython.address(sound_container.sample)
        sound_settings.volume = <Uint8>(sound_instance.volume * SDL_MIX_MAXVOLUME)
        sound_settings.volume_left = sound_settings.volume
        sound_settings.volume_right = sound_settings.volume
        sound_settings.loops_remaining = sound_instance.loops
        sound_settings.sound_priority = sound_instance.priority

        # If the sound is not set to loop, adjust the loop end position to the end of the sound
        if sound_settings.loops_remaining == 0:
            sound_settings.loop_end_pos = sound_container.sample.size

        # Apply simple linear pan/balance setting
        if self.state.callback_data.channels == 2:
            if sound_instance.pan < 0:
                sound_settings.volume_left = sound_settings.volume
                sound_settings.volume_right = <Uint8>(sound_instance.volume * SDL_MIX_MAXVOLUME * in_out_quad(sound_instance.pan + 1.0))
            else:
                sound_settings.volume_left = <Uint8>(sound_instance.volume * SDL_MIX_MAXVOLUME * in_out_quad(1.0 - sound_instance.pan))
                sound_settings.volume_right = sound_settings.volume

        # Fading (done at control rate; need to calculate the number of steps over which to fade in/out)
        sound_settings.fade_in_steps = sound_instance.fade_in * self.state.callback_data.seconds_to_bytes_factor // self.state.callback_data.bytes_per_control_point
        sound_settings.fade_out_steps = sound_instance.fade_out * self.state.callback_data.seconds_to_bytes_factor // self.state.callback_data.bytes_per_control_point
        sound_settings.fade_steps_remaining = sound_settings.fade_in_steps
        if sound_settings.fade_steps_remaining > 0:
            sound_settings.fading_status = fading_status_fading_in
        else:
            sound_settings.fading_status = fading_status_not_fading

        # Markers
        sound_settings.marker_count = sound_instance.marker_count
        g_array_set_size(sound_settings.markers, sound_settings.marker_count)
        for index in range(sound_instance.marker_count):
            g_array_insert_val_uint(sound_settings.markers,
                                    index,
                                    <guint>(sound_instance.markers[index]['time'] * self.state.callback_data.seconds_to_bytes_factor))

        # Set almost finished marker (calculate based on the end of the sound)
        if sound_instance.about_to_finish_time is None:
            sound_settings.about_to_finish_marker = no_marker
        elif sound_instance.about_to_finish_time > sound_container.duration:
            sound_settings.about_to_finish_marker = 0
        else:
            sound_settings.about_to_finish_marker = (sound_container.duration - sound_instance.about_to_finish_time) * self.state.callback_data.seconds_to_bytes_factor

        # If the sound has ducking settings, apply them
        if sound_instance.ducking is not None and sound_instance.ducking.track_bit_mask != 0:
            # To convert between the number of seconds and a buffer position (bytes), we need to
            # account for the sample rate (samples per second), the number of audio channels, and the
            # number of bytes per sample (all samples are 16 bits)
            sound_settings.sound_has_ducking = True
            sound_settings.ducking_stage = ducking_stage_delay
            sound_settings.ducking_settings.track_bit_mask = sound_instance.ducking.track_bit_mask
            sound_settings.ducking_settings.attack_start_pos = sound_instance.ducking.delay * self.state.callback_data.seconds_to_bytes_factor
            sound_settings.ducking_settings.attack_duration = sound_instance.ducking.attack * self.state.callback_data.seconds_to_bytes_factor
            sound_settings.ducking_settings.attenuation_volume = <Uint8>(sound_instance.ducking.attenuation * SDL_MIX_MAXVOLUME)
            sound_settings.ducking_settings.release_duration = sound_instance.ducking.release * self.state.callback_data.seconds_to_bytes_factor

            # Release point is relative to the end of the sound
            sound_settings.ducking_settings.release_start_pos = (sound_container.duration - sound_instance.ducking.release_point) * self.state.callback_data.seconds_to_bytes_factor
        else:
            # Sound does not have ducking, assign settings appropriately
            sound_settings.sound_has_ducking = False
            sound_settings.ducking_stage = ducking_stage_idle
            sound_settings.ducking_settings.track_bit_mask = 0
            sound_settings.ducking_settings.attack_start_pos = 0
            sound_settings.ducking_settings.attack_duration = 0
            sound_settings.ducking_settings.attenuation_volume = SDL_MIX_MAXVOLUME
            sound_settings.ducking_settings.release_duration = 0
            sound_settings.ducking_settings.release_start_pos = 0

        # Special handling is needed to start streaming for the specified sound at the correct location
        if sound_container.sample.type == sound_type_streaming:
            # Seek to the specified start position
            gst_element_seek_simple(sound_container.sample.data.stream.pipeline,
                                    GST_FORMAT_TIME,
                                    <GstSeekFlags>(GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_KEY_UNIT),
                                    sound_instance.start_at * GST_SECOND)
            with nogil:
                ret = gst_element_set_state(sound_container.sample.data.stream.pipeline, GST_STATE_PLAYING)

    cdef _set_player_playing(self, SoundPlayer *player, object sound_instance):
        """
        Sets the player status and sound settings to begin playing the sound instance
        Args:
            player: A pointer to the SoundPlayer on which to play the sound
            sound_instance: The sound instance to begin playing
        """
        if player == NULL or sound_instance is None:
            return

        # Setup the player to start playing the sound
        player.status = player_playing
        self._set_player_sound_settings(cython.address(player.current), sound_instance)

        # Send sound started notification
        send_sound_started_notification(player.number, player.current.sound_id, player.current.sound_instance_id, self.state)

        self.log.debug("Sound %s is set to begin playback on playlist track (loops=%d)",
                       sound_instance.name, sound_instance.loops)

    cdef _set_player_replacing(self, SoundPlayer *player, object sound_instance):
        """
        Sets the player status and sound settings to replace the currently playing sound with the
        sound instance.
        Args:
            player: A pointer to the SoundPlayer on which to replace the sound
            sound_instance: The sound instance to begin playing
        """
        if player == NULL or sound_instance is None:
            return

        # Set current sound to fade out quickly
        player.current.fade_out_steps = self.state.callback_data.quick_fade_steps
        player.current.fade_steps_remaining = self.state.callback_data.quick_fade_steps
        player.current.fading_status = fading_status_fading_out

        # Set the next sound to play immediately after the current one fades out
        player.status = player_replacing
        self._set_player_sound_settings(cython.address(player.next), sound_instance)

        # TODO: Figure out how to handle ducking when replacing an existing sound

    cdef int _get_player_playing_sound_instance(self, sound_instance):
        """
        Return the player currently playing the specified sound instance
        Args:
            sound_instance: The SoundInstance to find

        Returns:
            The sound player number currently playing the specified sound instance or -1 if the
            sound instance is not currently playing.
        """
        SDL_LockAudio()

        for i in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[i].status != player_idle and \
                            self.type_state.sound_players[i].current.sound_instance_id == sound_instance.id:
                SDL_UnlockAudio()
                return i

        SDL_UnlockAudio()
        return -1

    cdef Uint32 _fix_sample_frame_pos(self, Uint32 sample_pos, Uint8 bytes_per_sample, int channels):
        """
        Rounds up sample position to a sample frame boundary (audio distortion may occur if starting
        in the middle of a sample frame).

        Args:
            sample_pos: The sample position
            bytes_per_sample: The number of bytes per sample 
            channels: The number of audio channels

        Returns:
            sample_pos rounded to the nearest sample frame boundary
        """
        cdef int bytes_per_sample_frame = bytes_per_sample * channels
        return bytes_per_sample_frame * ceil(sample_pos / bytes_per_sample_frame)

    def get_playing_sound_instance_by_id(self, sound_instance_id):
        if sound_instance_id in self._playing_instances_by_id:
            return self._playing_instances_by_id[sound_instance_id]
        else:
            return None

    def get_status(self):
        """
        Get the current track status (status of all sound players on the track).
        Used for debugging and testing.
        Returns:
            A list of status dictionaries containing the current settings for each
            sound player.
        """
        SDL_LockAudio()
        status = []
        for player in range(self.type_state.sound_player_count):
            status.append({
                "player": player,
                "status": TrackStandard.player_status_to_text(<int>self.type_state.sound_players[player].status),
                "fading_status": TrackStandard.player_fading_status_to_text(<int>self.type_state.sound_players[player].current.fading_status),
                "volume": self.type_state.sound_players[player].current.volume,
                "sound_id": self.type_state.sound_players[player].current.sound_id,
                "sound_instance_id": self.type_state.sound_players[player].current.sound_instance_id,
                "priority": self.type_state.sound_players[player].current.sound_priority,
                "loops": self.type_state.sound_players[player].current.loops_remaining,
                "current_loop": self.type_state.sound_players[player].current.current_loop,
                "has_ducking": self.type_state.sound_players[player].current.sound_has_ducking,
                "sample_pos": self.type_state.sound_players[player].current.sample_pos,
                "loop_start_pos": self.type_state.sound_players[player].current.loop_start_pos,
                "loop_end_pos": self.type_state.sound_players[player].current.loop_end_pos
            })

            self.log.debug("Status - Player %d: Status=%s, Sound=%d, SoundInstance=%d"
                           "Priority=%d, Loops=%d, SamplePos=%d",
                           player,
                           TrackStandard.player_status_to_text(
                               self.type_state.sound_players[player].status),
                           self.type_state.sound_players[player].current.sound_id,
                           self.type_state.sound_players[player].current.sound_instance_id,
                           self.type_state.sound_players[player].current.sound_priority,
                           self.type_state.sound_players[player].current.loops_remaining,
                           self.type_state.sound_players[player].current.sample_pos)

        SDL_UnlockAudio()

        return status

    def get_sound_queue_count(self):
        """
        Gets the number of sounds currently in the track sound queue.
        Returns:
            Integer number of sounds currently in the track sound queue.
        """
        return len(self._sound_queue)

    def get_sound_players_in_use_count(self):
        """
        Gets the current count of sound players in use on the track.  Used for
        debugging and testing.
        Returns:
            Integer number of sound players currently in use on the track.
        """
        players_in_use_count = 0
        SDL_LockAudio()
        for i in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[i].status != player_idle:
                players_in_use_count += 1
        SDL_UnlockAudio()
        return players_in_use_count

    def sound_is_playing(self, sound not None):
        """Returns whether or not the specified sound is currently playing on the track"""
        SDL_LockAudio()
        for i in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[i].status != player_idle and \
                            self.type_state.sound_players[i].current.sound_id == sound.id:
                SDL_UnlockAudio()
                return True

        SDL_UnlockAudio()
        return False

    def sound_instance_is_playing(self, sound_instance not None):
        """Returns whether or not the specified sound instance is currently playing on the track"""
        SDL_LockAudio()
        for i in range(self.type_state.sound_player_count):
            if self.type_state.sound_players[i].status != player_idle and \
                            self.type_state.sound_players[i].current.sound_instance_id == sound_instance.id:
                SDL_UnlockAudio()
                return True

        SDL_UnlockAudio()
        return False

    def sound_is_in_queue(self, sound not None):
        """Returns whether or not an instance of the specified sound is currently in the queue"""
        for sound_instance in self._sound_queue:
            if sound_instance.sound.id == sound.id:
                return True

        return False

    def sound_instance_is_in_queue(self, sound_instance not None):
        """Returns whether or not the specified sound instance is currently in the queue"""
        return sound_instance in self._sound_queue

    @staticmethod
    def player_status_to_text(int status):
        """
        Converts a sound player status value into an equivalent text string.  Used for testing
        purposes only.
        Args:
            status: Integer sound player status value

        Returns:
            string containing the equivalent status text
        """
        status_values = {
            player_idle: "idle",
            player_pending: "pending",
            player_replacing: "replacing",
            player_playing: "playing",
            player_finished: "finished",
            player_stopping: "stopping",
        }

        try:
            return status_values.get(status)
        except KeyError:
            return "unknown"

    @staticmethod
    def player_fading_status_to_text(int fading_status):
        """
        Converts a sound player fading status value into an equivalent text string.  Used for
        testing purposes only.
        Args:
            fading_status: Integer sound player fading status value

        Returns:
            string containing the equivalent fading status text
        """
        fading_status_values = {
            fading_status_not_fading: "not fading",
            fading_status_fading_in: "fade in",
            fading_status_fading_out: "fade out",
        }

        try:
            return fading_status_values.get(fading_status)
        except KeyError:
            return "unknown"


# ---------------------------------------------------------------------------
#    Global C functions designed to be called from the static audio callback
#    function (these functions do not use the GIL).
#
#    Note: Because these functions are only called from the audio callback
#    function, we do not need to lock and unlock the mutex in these functions
#    (locking/unlocking of the mutex is already performed in the audio
#    callback function.
# ---------------------------------------------------------------------------

    @staticmethod
    cdef void mix_playing_sounds(TrackState *track, Uint32 buffer_length, AudioCallbackData *callback_data) nogil:
        """
        Mixes any sounds that are playing on the specified standard track into the specified audio buffer.
        Args:
            track: A pointer to the TrackState data structure for the track
            buffer_length: The length of the output buffer (in bytes)
            callback_data: The audio callback data structure
        Notes:
            Notification messages are generated.
        """
        cdef TrackState *target_track
        cdef TrackStandardState *standard_track
        cdef SoundPlayer *player
        cdef int player_num
        cdef int track_num
        cdef int marker_id
        cdef bint end_of_sound
        cdef bint ducking_is_active

        if track == NULL or track.type_state == NULL:
            return

        standard_track = <TrackStandardState*>track.type_state

        # Setup local variables
        cdef Uint32 buffer_bytes_remaining
        cdef Uint32 current_chunk_bytes
        cdef Uint32 track_buffer_pos
        cdef Uint8 control_point
        cdef float progress

        # Loop over track sound players
        for player_num in range(standard_track.sound_player_count):

            player = cython.address(standard_track.sound_players[player_num])

            # If the player is idle, there is nothing to do so move on to the next player
            if player == NULL or player.status == player_idle:
                continue

            # Set flag indicating there is at least some activity on the track (it is active)
            track.active = True

            end_of_sound = False
            track_buffer_pos = 0
            control_point = 0
            buffer_bytes_remaining = buffer_length

            # Loop over output buffer at control rate
            while buffer_bytes_remaining > 0:

                # Determine the number of bytes to process in the current chunk
                current_chunk_bytes = min(buffer_bytes_remaining, callback_data.bytes_per_control_point)

                # Calculate volume of chunk (handle fading)
                if player.current.fading_status == fading_status_fading_in:
                    volume = <Uint8> (in_out_quad((player.current.fade_in_steps - player.current.fade_steps_remaining) / player.current.fade_in_steps) * player.current.volume)
                    player.current.fade_steps_remaining -= 1
                    if player.current.fade_steps_remaining == 0:
                        player.current.fading_status = fading_status_not_fading
                elif player.current.fading_status == fading_status_fading_out:
                    volume = <Uint8> (in_out_quad(player.current.fade_steps_remaining / player.current.fade_out_steps) * player.current.volume)
                    player.current.fade_steps_remaining -= 1
                else:
                    volume = player.current.volume

                # Copy samples for chunk to output buffer and apply volume
                if player.current.sample.type == sound_type_memory:
                    end_of_sound = get_memory_sound_samples(cython.address(player.current), current_chunk_bytes,
                                                            track.buffer + track_buffer_pos, callback_data.channels,
                                                            volume, track, player_num)
                elif player.current.sample.type == sound_type_streaming:
                    end_of_sound = get_streaming_sound_samples(cython.address(player.current), current_chunk_bytes,
                                                               track.buffer + track_buffer_pos, callback_data.channels,
                                                               volume, track, player_num)

                # Process sound ducking (if applicable)
                if player.current.sound_has_ducking:
                    ducking_is_active = False

                    # Determine control point ducking stage and calculate control point (test stages in reverse order)
                    if player.current.sample_pos >= player.current.ducking_settings.release_start_pos + player.current.ducking_settings.release_duration:
                        # Ducking finished
                        g_array_set_val_uint8(player.current.ducking_control_points, control_point, SDL_MIX_MAXVOLUME)

                    elif player.current.sample_pos >= player.current.ducking_settings.release_start_pos:
                        # Ducking release stage
                        ducking_is_active = True
                        progress = (player.current.sample_pos - player.current.ducking_settings.release_start_pos) / player.current.ducking_settings.release_duration
                        g_array_set_val_uint8(player.current.ducking_control_points,
                                              control_point,
                                              lerpU8(in_out_quad(progress),
                                                     player.current.ducking_settings.attenuation_volume,
                                                     SDL_MIX_MAXVOLUME
                                                     )
                                              )

                    elif player.current.sample_pos >= player.current.ducking_settings.attack_start_pos + player.current.ducking_settings.attack_duration:
                        # Ducking hold state
                        ducking_is_active = True
                        g_array_set_val_uint8(player.current.ducking_control_points, control_point, player.current.ducking_settings.attenuation_volume)

                    elif player.current.sample_pos >= player.current.ducking_settings.attack_start_pos:
                        # Ducking attack stage
                        ducking_is_active = True
                        progress = (player.current.sample_pos - player.current.ducking_settings.attack_start_pos) / player.current.ducking_settings.attack_duration
                        g_array_set_val_uint8(player.current.ducking_control_points,
                                              control_point,
                                              lerpU8(in_out_quad(progress),
                                                     SDL_MIX_MAXVOLUME,
                                                     player.current.ducking_settings.attenuation_volume
                                                     )
                                              )

                    else:
                        # Ducking delay stage
                        g_array_set_val_uint8(player.current.ducking_control_points,
                                              control_point,
                                              SDL_MIX_MAXVOLUME)

                    # Apply ducking to target track(s) (when applicable)
                    if ducking_is_active:
                        for track_num in range(callback_data.track_count):
                            target_track = <TrackState*>callback_data.tracks[track_num]
                            if (1 << track_num) & player.current.ducking_settings.track_bit_mask:
                                target_track.ducking_is_active = True
                                g_array_set_val_uint8(target_track.ducking_control_points,
                                                      control_point,
                                                      min(
                                                          g_array_index_uint8(
                                                              target_track.ducking_control_points,
                                                              control_point),
                                                          g_array_index_uint8(player.current.ducking_control_points,
                                                                              control_point)
                                                      ))

                    # TODO: Hold sound processing until ducking has finished
                    # It is possible to have the ducking release finish after the sound has stopped.  In that
                    # case, silence should be generated until the ducking is done.

                # Process markers (do any markers fall in the current chunk?)
                # Note: the current sample position has already been incremented when the sample data was received so
                # we need to look backwards from the current position to determine if marker falls in chunk window.

                # About to finish marker
                if player.current.about_to_finish_marker != no_marker:
                    if player.current.sample_pos - current_chunk_bytes <= player.current.about_to_finish_marker < player.current.sample_pos:
                        # Marker is in window, send notification
                        send_sound_about_to_finish_notification(player_num,
                                                                player.current.sound_id,
                                                                player.current.sound_instance_id,
                                                                track)
                    # Special check if buffer wraps back around to the beginning of the sample
                    if not end_of_sound and player.current.sample_pos - current_chunk_bytes < 0 and player.current.about_to_finish_marker < player.current.sample_pos:
                        # Marker is in window, send notification
                        send_sound_about_to_finish_notification(player_num,
                                                                player.current.sound_id,
                                                                player.current.sound_instance_id,
                                                                track)

                # User-defined markers
                for marker_id in range(player.current.marker_count):
                    if player.current.sample_pos - current_chunk_bytes <= g_array_index_uint(player.current.markers, marker_id) < player.current.sample_pos:
                        # Marker is in window, send notification
                        send_sound_marker_notification(player_num,
                                                       player.current.sound_id,
                                                       player.current.sound_instance_id,
                                                       track,
                                                       marker_id)
                    # Special check if buffer wraps back around to the beginning of the sample
                    if not end_of_sound and player.current.sample_pos - current_chunk_bytes < 0 and g_array_index_uint(player.current.markers, marker_id) < player.current.sample_pos:
                        # Marker is in window, send notification
                        send_sound_marker_notification(player_num,
                                                       player.current.sound_id,
                                                       player.current.sound_instance_id,
                                                       track,
                                                       marker_id)

                # Check if sound is finished due to a fade out completing
                if player.current.fading_status == fading_status_fading_out and player.current.fade_steps_remaining == 0:
                    end_of_sound = True

                # Sound finished processing
                if end_of_sound:
                    send_sound_stopped_notification(player_num, player.current.sound_id, player.current.sound_instance_id, track)

                    # End of sound behavior depends upon player status
                    if player.status == player_replacing:
                        # Replacing the current sound with a new one: copy sound player settings from next sound to current
                        player.current.sample = player.next.sample
                        player.current.sample_pos = player.next.sample_pos
                        player.current.loop_start_pos = player.next.loop_start_pos
                        player.current.loop_end_pos = player.next.loop_end_pos
                        player.current.current_loop = player.next.current_loop
                        player.current.sound_id = player.next.sound_id
                        player.current.sound_instance_id = player.next.sound_instance_id
                        player.current.volume = player.next.volume
                        player.current.loops_remaining = player.next.loops_remaining
                        player.current.sound_priority = player.next.sound_priority
                        player.current.fade_in_steps = player.next.fade_in_steps
                        player.current.fade_out_steps = player.next.fade_out_steps
                        player.current.fade_steps_remaining = player.next.fade_steps_remaining
                        player.current.fading_status = player.next.fading_status
                        player.current.about_to_finish_marker = player.next.about_to_finish_marker

                        player.current.marker_count = player.next.marker_count
                        g_array_set_size(player.current.markers, player.current.marker_count)
                        for marker_id in range(player.next.marker_count):
                            g_array_insert_val_uint(player.current.markers,
                                                    marker_id,
                                                    g_array_index_uint(player.next.markers, marker_id))

                        if player.next.sound_has_ducking:
                            player.current.sound_has_ducking = True
                            player.current.ducking_stage = ducking_stage_delay
                            player.current.ducking_settings.track_bit_mask = player.next.ducking_settings.track_bit_mask
                            player.current.ducking_settings.attack_start_pos = player.next.ducking_settings.attack_start_pos
                            player.current.ducking_settings.attack_duration = player.next.ducking_settings.attack_duration
                            player.current.ducking_settings.attenuation_volume = player.next.ducking_settings.attenuation_volume
                            player.current.ducking_settings.release_start_pos = player.next.ducking_settings.release_start_pos
                            player.current.ducking_settings.release_duration = player.next.ducking_settings.release_duration
                        else:
                            player.current.sound_has_ducking = False
                            player.current.ducking_stage = ducking_stage_idle
                            player.current.ducking_settings.track_bit_mask = 0
                            player.current.ducking_settings.attack_start_pos = 0
                            player.current.ducking_settings.attack_duration = 0
                            player.current.ducking_settings.attenuation_volume = SDL_MIX_MAXVOLUME
                            player.current.ducking_settings.release_start_pos = 0
                            player.current.ducking_settings.release_duration = 0

                        # Send sound started notification
                        send_sound_started_notification(player_num, player.current.sound_id, player.current.sound_instance_id, track)
                        player.status = player_playing
                        sound_finished = False
                    else:
                        player.status = player_idle
                        break

                # Move to next chunk
                buffer_bytes_remaining -= current_chunk_bytes
                track_buffer_pos += current_chunk_bytes
                control_point += 1

cdef bint get_memory_sound_samples(SoundSettings *sound, Uint32 length, Uint8 *output_buffer, int channels,
                                   Uint8 volume, TrackState *track, int player_num) nogil:
    """
    Retrieves the specified number of bytes from the source sound memory buffer and mixes them into
    the track output buffer at the specified volume.

    Args:
        sound: A pointer to a SoundSettings struct (contains all sound state and settings to play the sound)
        length: The number of samples to retrieve and place in the output buffer
        output_buffer: The output buffer
        channels: The number of channels in the output buffer (1 = mono, 2 = stereo)
        volume: The volume to apply to the output buffer (fixed for the duration of this method)
        track: A pointer to the TrackState struct for the current track
        player_num: The sound player number currently playing the sound (used for notification messages)

    Returns:
        True if sound is finished, False otherwise
    """
    if sound == NULL or output_buffer == NULL:
        return True

    cdef Uint32 samples_remaining_to_output = length
    cdef Uint32 samples_remaining_in_sound
    cdef Uint32 loop_end_pos = sound.loop_end_pos
    cdef Uint32 buffer_pos = 0
    cdef Uint8 *sound_buffer = <Uint8*>sound.sample.data.memory.data
    if sound_buffer == NULL:
        return True

    # Make sure the sound sample position is not starting after the end of the loop position. If so, adjust the
    # end of the loop position to be the end of the sound (keeps the sound player from running past the end of
    # the sound buffer and into other parts of memory creating noise). This should only ever happen if the sound
    # is started after the end of the loop position.
    if sound.sample_pos >= loop_end_pos:
        loop_end_pos = sound.sample.data.memory.size

    # Loop while there are still samples remaining to be output
    while samples_remaining_to_output > 0:

        # Determine how many samples are remaining in the sound buffer before the end of the sound (the
        # current loop position)
        samples_remaining_in_sound = loop_end_pos - sound.sample_pos

        # Determine if we are consuming the entire remaining sound buffer, or just a portion of it
        if samples_remaining_to_output < samples_remaining_in_sound:

            # We are not consuming the entire streaming buffer.  There will still be buffer data remaining for the next call.
            if channels == 2 and False:
                # disabled for now
                Track.mix_audio_stereo(output_buffer + buffer_pos,
                                       <Uint8*>sound.sample.data.memory.data + sound.sample_pos,
                                       samples_remaining_to_output,
                                       sound.volume_left,
                                       sound.volume_right)
            else:
                SDL_MixAudioFormat(output_buffer + buffer_pos, <Uint8*>sound.sample.data.memory.data + sound.sample_pos, track.callback_data.format, samples_remaining_to_output, volume)

            # Update buffer position pointers
            sound.sample_pos += samples_remaining_to_output

            # Sound is not finished, but the output buffer has been filled
            return False
        else:
            # Entire sound buffer consumed. Mix in remaining samples
            if channels == 2 and False:
                # disabled for now
                Track.mix_audio_stereo(output_buffer + buffer_pos,
                                       <Uint8*>sound.sample.data.memory.data + sound.sample_pos,
                                       samples_remaining_to_output,
                                       sound.volume_left,
                                       sound.volume_right)
            else:
                SDL_MixAudioFormat(output_buffer + buffer_pos, <Uint8*>sound.sample.data.memory.data + sound.sample_pos, track.callback_data.format, samples_remaining_in_sound, volume)

            # Update buffer position pointers/samples remaining to place in the output buffer
            samples_remaining_to_output -= samples_remaining_in_sound
            sound.sample_pos += samples_remaining_in_sound
            buffer_pos += samples_remaining_in_sound

        # Check if we are at the end of the source sample buffer (loop if applicable)
        if sound.sample_pos >= loop_end_pos:
            if sound.loops_remaining > 0:
                # At the end and still loops remaining, loop back to the beginning of the loop
                sound.loops_remaining -= 1
                sound.sample_pos = sound.loop_start_pos
                sound.current_loop += 1
                send_sound_looping_notification(player_num, sound.sound_id, sound.sound_instance_id, track)

                # If the sound is on its last loop, set the loop end position to be the end of the sound
                if sound.loops_remaining == 0:
                    sound.loop_end_pos = sound.sample.data.memory.size
                    loop_end_pos = sound.loop_end_pos

            elif sound.loops_remaining == 0:
                # At the end and not looping, the sample has finished playing (return True for end of sound)
                return True

            else:
                # Looping infinitely, loop back to the beginning of the loop
                sound.sample_pos = sound.loop_start_pos
                sound.current_loop += 1
                send_sound_looping_notification(player_num, sound.sound_id, sound.sound_instance_id, track)

    return False

cdef bint get_streaming_sound_samples(SoundSettings *sound, Uint32 length, Uint8 *output_buffer, int channels,
                                      Uint8 volume, TrackState *track, int player_num) nogil:
    """
    Retrieves the specified number of bytes from the source sound streaming buffer and mixes them
    into the track output buffer at the specified volume.

    Args:
        sound: A pointer to a SoundSettings struct (contains all sound state and settings to play the sound)
        length: The number of samples to retrieve and place in the output buffer
        output_buffer: The output buffer
        channels: The number of channels in the output buffer (1 = mono, 2 = stereo)
        volume: The volume to apply to the output buffer (fixed for the duration of this method)
        track: A pointer to the TrackState struct for the current track
        player_num: The sound player number currently playing the sound (used for notification messages)

    Returns:
        True if sound is finished, False otherwise

    Notes:
        The important thing to consider about retrieving samples from the streaming sound source
        is the buffer size used by SDL2 (output) and GStreamer (input) may be very different. A
        buffer is "pulled" synchronously from the streaming source and is held until it is
        completely consumed.  At which point either the sound ends if the source reports is at the
        end of stream (eos), or another buffer is pulled.
    """
    if sound == NULL or output_buffer == NULL or sound.sample.data.stream.pipeline == NULL:
        return True

    cdef Uint32 samples_remaining_to_output = length
    cdef Uint32 samples_remaining_in_map
    cdef Uint32 buffer_pos = 0

    while samples_remaining_to_output > 0:

        # Copy any samples remaining in the streaming buffer
        if sound.sample.data.stream.map_contains_valid_sample_data:
            samples_remaining_in_map = sound.sample.data.stream.map_info.size - sound.sample.data.stream.map_buffer_pos

            # Determine if we are consuming the entire buffer of streaming samples, or just a portion of it
            if samples_remaining_to_output < samples_remaining_in_map:
                # We are not consuming the entire streaming buffer.  There will still be buffer data remaining for the next call.
                if channels == 2 and False:
                    # disabled for now
                    Track.mix_audio_stereo(output_buffer + buffer_pos,
                                           sound.sample.data.stream.map_info.data + sound.sample.data.stream.map_buffer_pos,
                                           samples_remaining_to_output,
                                           sound.volume_left,
                                           sound.volume_right)
                else:
                    SDL_MixAudioFormat(output_buffer + buffer_pos,
                                       sound.sample.data.stream.map_info.data + sound.sample.data.stream.map_buffer_pos,
                                       track.callback_data.format, samples_remaining_to_output, volume)

                # Update buffer position pointers
                sound.sample.data.stream.map_buffer_pos += samples_remaining_to_output
                sound.sample_pos += samples_remaining_to_output

                # Sound is not finished, but the output buffer has been filled
                return False
            else:
                # Entire buffer of leftover samples consumed.  Free the buffer resources to prepare for next call
                if channels == 2 and False:
                    # disabled for now
                    Track.mix_audio_stereo(output_buffer + buffer_pos,
                                           sound.sample.data.stream.map_info.data + sound.sample.data.stream.map_buffer_pos,
                                           samples_remaining_to_output,
                                           sound.volume_left,
                                           sound.volume_right)
                else:
                    SDL_MixAudioFormat(output_buffer + buffer_pos,
                                       sound.sample.data.stream.map_info.data + sound.sample.data.stream.map_buffer_pos,
                                       track.callback_data.format, samples_remaining_in_map, volume)

                # Update buffer position pointers/samples remaining to place in the output buffer
                samples_remaining_to_output -= samples_remaining_in_map
                sound.sample_pos += samples_remaining_in_map
                buffer_pos += samples_remaining_in_map

                # Done with the streaming buffer, release references to it
                gst_buffer_unmap(sound.sample.data.stream.buffer, &sound.sample.data.stream.map_info)
                gst_sample_unref(sound.sample.data.stream.sample)

                sound.sample.data.stream.buffer = NULL
                sound.sample.data.stream.sample = NULL
                sound.sample.data.stream.map_buffer_pos = 0
                sound.sample.data.stream.map_contains_valid_sample_data = 0

        # Check for eos (end of stream)
        if g_object_get_bool(sound.sample.data.stream.sink, "eos"):

            # At the end of the stream - check if sound should loop or end
            if sound.loops_remaining > 0:
                # At the end and still loops remaining, loop back to the beginning
                sound.loops_remaining -= 1
                sound.sample_pos = 0
                sound.current_loop += 1
                send_sound_looping_notification(player_num, sound.sound_id, sound.sound_instance_id, track)

            elif sound.loops_remaining == 0:
                # At the end and not looping, the sample has finished playing (return True for end of sound)
                return True

            else:
                # Looping infinitely, loop back to the beginning
                sound.sample_pos = 0
                sound.current_loop += 1
                send_sound_looping_notification(player_num, sound.sound_id, sound.sound_instance_id, track)

            # Seek back to the beginning of the sound's source file
            gst_element_seek_simple(sound.sample.data.stream.pipeline, GST_FORMAT_TIME, <GstSeekFlags>(GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_KEY_UNIT), 0)

        # Retrieve the next buffer from the streaming pipeline
        sound.sample.data.stream.sample = c_appsink_pull_sample(sound.sample.data.stream.sink)

        if sound.sample.data.stream.sample == NULL:
            sound.sample.data.stream.null_buffer_count += 1

            # If we've received too many consecutive null buffers, end the sound
            if sound.sample.data.stream.null_buffer_count > CONSECUTIVE_NULL_STREAMING_BUFFER_LIMIT:
                return True
        else:
            sound.sample.data.stream.null_buffer_count = 0
            sound.sample.data.stream.buffer = gst_sample_get_buffer(sound.sample.data.stream.sample)

            if gst_buffer_map(sound.sample.data.stream.buffer, &sound.sample.data.stream.map_info, GST_MAP_READ):
                sound.sample.data.stream.map_contains_valid_sample_data = 1
                sound.sample.data.stream.map_buffer_pos = 0
            else:
                sound.sample.data.stream.map_contains_valid_sample_data = 0
                sound.sample.data.stream.map_buffer_pos = 0
                gst_sample_unref(sound.sample.data.stream.sample)
                sound.sample.data.stream.sample = NULL

    # The sound has not finished playing, but the output buffer has been filled
    return False



# How to add a test:
# Copy this file
# Rename TestTemplate to TestWhatever in line 9
# Rename machine path and config file in lines 11 and 14

from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestTemplate(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/test_template'

    def get_config_file(self):
        return 'test_template.yaml'

    def test_something(self):
        pass



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase
from unittest.mock import MagicMock


class TestKeyboard(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/keyboard'

    def get_config_file(self):
        return 'test_keyboard.yaml'

    def press(self, key_string, mods=None):
        if mods:
            if not isinstance(mods, list):
                mods = [mods]
        else:
            mods = list()

        self.mc.keyboard._on_keyboard_down(None, (None, key_string.lower()),
                                           None,
                                           mods)
        # self.advance_time()

    def release(self, key_string):
        self.mc.keyboard._on_keyboard_up(None, (None, key_string.lower()))
        # self.advance_time()

    def test_switch(self):
        self.press('a')
        bcp_command = ('switch', None, {'name': 'switch_a', 'state': 1})
        self.assertIn(bcp_command, self.sent_bcp_commands)

        self.release('a')
        bcp_command = ('switch', None, {'name': 'switch_a', 'state': 0})
        self.assertIn(bcp_command, self.sent_bcp_commands)

        self.sent_bcp_commands = list()

        # When a key is held down, kivy keeps firing the same event over and
        # over each frame, so make sure that we only send one event
        self.press('a')
        self.assertEqual(len(self.sent_bcp_commands), 1)

        self.press('a')
        self.press('a')
        self.press('a')
        self.press('a')
        self.press('a')
        self.press('a')
        self.assertEqual(len(self.sent_bcp_commands), 1)

    def test_toggle_switch(self):
        # BCP spec sends state -1 to tell MPF to flip the switch state. -1 is
        # used so mpf-mc doesn't have to maintain of state which means it can't
        # get out of sync. :)
        self.press('b')
        bcp_command = ('switch', None, {'name': 'switch_b', 'state': -1})
        self.assertIn(bcp_command, self.sent_bcp_commands)

        self.release('b')

        # make sure releasing doesn't send it again
        found = 0
        for x in self.sent_bcp_commands:
            if x == ('switch', None, {'name': 'switch_b', 'state': -1}):
                found += 1

        self.assertEqual(found, 1)

    def test_inverted_switch(self):
        self.press('c')
        bcp_command = ('switch', None, {'name': 'switch_c', 'state': 0})
        self.assertIn(bcp_command, self.sent_bcp_commands)

        self.release('c')
        bcp_command = ('switch', None, {'name': 'switch_c', 'state': 1})
        self.assertIn(bcp_command, self.sent_bcp_commands)

    def test_mpf_event(self):
        self.mc.bcp_processor.enabled = True

        self.press('d')

        bcp_command = ('trigger', None, {'name': 'event_d'})
        self.assertIn(bcp_command, self.sent_bcp_commands)

    def test_mpf_event_with_params(self):
        self.mc.bcp_processor.enabled = True

        self.press('e')

        bcp_command = ('trigger', None, {'mission': 'pinball', 'foo': 'bar',
                                         'name': 'event_e'})
        self.assertIn(bcp_command, self.sent_bcp_commands)

    def test_mc_event(self):
        self.callback = MagicMock()

        self.mc.events.add_handler('event_f', self.callback)

        self.press('f')
        self.advance_time()
        self.callback.assert_called_once_with()

    def test_mc_event_with_params(self):
        self.callback = MagicMock()

        self.mc.events.add_handler('event_g', self.callback)

        self.press('g')
        self.advance_time()
        self.callback.assert_called_once_with(foo='bar', mission='pinball')

    def test_mod_key_with_dash(self):
        self.press('a', 'shift')

        bcp_command = ('switch', None, {'name': 'shift_a', 'state': 1})
        self.assertIn(bcp_command, self.sent_bcp_commands)

    def test_mod_key_with_plus(self):
        self.press('b', 'shift')

        bcp_command = ('switch', None, {'name': 'shift_b', 'state': 1})
        self.assertIn(bcp_command, self.sent_bcp_commands)

    def test_multiple_mod_keys(self):
        self.press('c', ['shift', 'ctrl'])
        self.release('c')

        bcp_command = ('switch', None, {'name': 'shift_ctrl_c', 'state': 1})
        self.assertIn(bcp_command, self.sent_bcp_commands)

        self.sent_bcp_commands = list()

        # swap the order of the mod keys
        self.press('c', ['ctrl', 'shift'])

        bcp_command = ('switch', None, {'name': 'shift_ctrl_c', 'state': 1})
        self.assertIn(bcp_command, self.sent_bcp_commands)

    def test_number_key(self):
        self.press('1')
        bcp_command = ('switch', None, {'name': 'switch_1', 'state': 1})
        self.assertIn(bcp_command, self.sent_bcp_commands)

    def test_period(self):
        self.press('.')
        bcp_command = ('switch', None, {'name': 'switch_period', 'state': 1})
        self.assertIn(bcp_command, self.sent_bcp_commands)

    def test_slash(self):
        self.press('/')
        bcp_command = ('switch', None, {'name': 'switch_slash', 'state': 1})
        self.assertIn(bcp_command, self.sent_bcp_commands)



from unittest.mock import MagicMock

from mpfmc._version import __version__
from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestBcpProcessor(MpfMcTestCase):

    def __init__(self, *args):
        super().__init__(*args)

    def get_machine_path(self):
        return 'tests/machine_files/bcp'

    def get_config_file(self):
        return 'test_bcp_processor.yaml'

    def test_game_flow(self):
        self.send('hello',
                  version='1.1',
                  controller_version=__version__,
                  controller_name='Mission Pinball Framework')
        self.advance_time()

        response = ('hello', None, {'version': '1.1'})
        self.assertIn(response, self.sent_bcp_commands)

        self.send('mode_start',
                  name='tilt',
                  priority='10000')
        self.advance_time()
        self.send('mode_start',
                  name='attract',
                  priority='10')
        self.advance_time()

        self.send('mode_start',
                  name='game',
                  priority='20')
        self.advance_time()

        self.send('player_added',
                  player_num='1')
        self.advance_time()

        self.send('player_score',
                  value='0',
                  player_num='1',
                  prev_value='0',
                  change='False')
        self.advance_time()

        self.send('player_turn_start',
                  player_num='1')
        self.advance_time()

        self.send('player_variable',
                  name='ball',
                  value='1',
                  player_num='1',
                  prev_value='0',
                  change='True')
        self.advance_time()

        self.send('mode_stop',
                  name='attract')
        self.advance_time()

    def test_machine_variable(self):
        self.send('machine_variable',
                  value='FREE PLAY',
                  name='credits_string')
        self.assertEqual(self.mc.machine_vars['credits_string'], 'FREE PLAY')

        self.callback = MagicMock()
        self.mc.events.add_handler('machine_var_foo', self.callback)

        self.send('machine_variable',
                  value='0',
                  name='foo')
        self.assertEqual(self.mc.machine_vars['foo'], '0')
        self.advance_time()
        self.callback.assert_called_with(value='0', prev_value=None,
                                         change=True)

        self.callback.reset_mock()
        self.send('machine_variable',
                  value='10',
                  name='foo',
                  prev_value='0',
                  change='10')
        self.assertEqual(self.mc.machine_vars['foo'], '10')
        self.advance_time()
        self.callback.assert_called_with(value='10', prev_value='0',
                                         change='10')



from importlib import import_module
import multiprocessing


def _start_mc(mpf_path, machine_path, args):
    module = import_module('mpfmc.commands.mc')
    module.Command(mpf_path, machine_path, args)


def _start_imc(mpf_path, machine_path, args):
    from mpfmc.tools.interactive_mc.imc import InteractiveMc
    InteractiveMc(mpf_path, machine_path, args).run()


class Command:

    """Command which runs imc and mc."""

    def __init__(self, mpf_path, machine_path, args):
        """Run imc and mc."""

        mc = multiprocessing.Process(target=_start_mc,
                                     args=(mpf_path, machine_path, args))
        sb = multiprocessing.Process(target=_start_imc,
                                     args=(mpf_path, machine_path, args))

        multiprocessing.set_start_method('spawn')

        mc.start()
        sb.start()

        mc.join()
        sb.join()


def get_command():
    return 'imc', Command



"""DMD (hardware device)."""
import struct

from kivy.graphics.instructions import Callback
from kivy.uix.effectwidget import EffectWidget

from kivy.clock import Clock
from kivy.graphics.fbo import Fbo
from kivy.graphics.opengl import glReadPixels, GL_RGB, GL_UNSIGNED_BYTE
from kivy.graphics.texture import Texture

from mpfmc.effects.gain import GainEffect
from mpfmc.effects.flip_vertical import FlipVerticalEffect
from mpfmc.effects.gamma import GammaEffect

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc


class DmdBase:
    """Base class for DMD devices."""

    dmd_name_string = 'DMD'

    def __init__(self, mc: "MpfMc", name: str, config: dict) -> None:
        """Initialise DMD."""

        self.mc = mc
        self.name = name

        self.mc.log.info('Initializing DMD')

        self.config = self._get_validated_config(config)

        self.source = self.mc.displays[self.config['source_display']]
        self.prev_data = None
        self._dirty = True

        # put the widget canvas on a Fbo
        texture = Texture.create(size=self.source.size, colorfmt='rgb')
        self.fbo = Fbo(size=self.source.size, texture=texture)

        self.effect_widget = EffectWidget()

        effect_list = list()
        effect_list.append(FlipVerticalEffect())

        if self.config['brightness'] != 1.0:
            if not 0.0 <= self.config['brightness'] <= 1.0:
                raise ValueError("DMD brightness value should be between 0.0 "
                                 "and 1.0. Yours is {}".format(self.config['brightness']))

            effect_list.append(GainEffect(gain=self.config['brightness']))

        if self.config['gamma'] != 1.0:
            effect_list.append(GammaEffect(gamma=self.config['gamma']))

        self.effect_widget.effects = effect_list
        self.effect_widget.size = self.source.size

        self.fbo.add(self.effect_widget.canvas)

        with self.source.canvas:
            self.callback = Callback(self._trigger_rendering)

        self._set_dmd_fps()

    def _trigger_rendering(self, *args):
        del args
        self._dirty = True

    def _get_validated_config(self, config: dict) -> dict:
        raise NotImplementedError

    def _set_dmd_fps(self) -> None:
        # fps is the rate that the connected client requested. We'll use the
        # lower of the two
        mc_fps = self.config['fps']

        if mc_fps == 0:
            # pylint: disable-msg=protected-access
            mc_fps = Clock._max_fps

        # pylint: disable-msg=protected-access
        if mc_fps > Clock._max_fps:
            self.mc.log.warning("%s fps is higher than mpf-mc fps. "
                                "Will use mpf-mc fps setting for the DMD.",
                                DmdBase.dmd_name_string)
            # pylint: disable-msg=protected-access
            fps = Clock._max_fps
            update = 0
        # pylint: disable-msg=protected-access
        elif Clock._max_fps > mc_fps > 0:
            fps = mc_fps
            update = 1 / fps
        else:
            # pylint: disable-msg=protected-access
            fps = Clock._max_fps
            update = 0

        Clock.schedule_interval(self.tick, update)
        self.mc.log.info("Setting %s to %sfps",
                         DmdBase.dmd_name_string, fps)

    def tick(self, *args) -> None:
        """Draw image for DMD and send it."""
        del args
        # run this at the end of the tick to make sure all kivy bind callbacks have executed
        if self._dirty:
            Clock.schedule_once(self._render, -1)

    def _render(self, dt):
        del dt
        self._dirty = False
        widget = self.source
        fbo = self.fbo

        # detach the widget from the parent
        parent = widget.parent
        if parent and hasattr(parent, "remove_display_source"):
            parent.remove_display_source(widget)

        # clear the fbo background
        fbo.bind()
        fbo.clear_buffer()
        fbo.release()

        self.effect_widget.add_widget(widget.container)

        fbo.draw()

        fbo.bind()
        data = glReadPixels(0, 0, widget.native_size[0], widget.native_size[1],
                            GL_RGB, GL_UNSIGNED_BYTE)
        fbo.release()

        self.effect_widget.remove_widget(widget.container)

        # reattach to the parent
        if parent and hasattr(parent, "add_display_source"):
            parent.add_display_source(widget)

        if not self.config['only_send_changes'] or self.prev_data != data:
            self.prev_data = data
            self.send(data)

    def send(self, data: bytes) -> None:
        """Send data to DMD via BCP."""
        raise NotImplementedError


class Dmd(DmdBase):
    """Monochrome DMD."""

    def _get_validated_config(self, config: dict) -> dict:
        return self.mc.config_validator.validate_config('dmds', config)

    @classmethod
    def _convert_to_single_bytes(cls, data, config: dict) -> bytes:
        new_data = bytearray()
        loops = 0
        config.setdefault('luminosity', (.299, .587, .114))
        luminosity = config['luminosity']

        for r, g, b in struct.iter_unpack('BBB', data):
            loops += 1
            try:
                pixel_weight = ((r * luminosity[0]) + (g * luminosity[1]) + (b * luminosity[2])) / 255.
                new_data.append(int(round(pixel_weight * 15)))

            except ValueError:
                raise ValueError(loops, r, g, b)

        return bytes(new_data)

    def send(self, data: bytes) -> None:
        """Send data to DMD via BCP."""
        data = self._convert_to_single_bytes(data, self.config)

        self.mc.bcp_processor.send('dmd_frame', rawbytes=data, name=self.name)


class RgbDmd(DmdBase):
    """RGB DMD."""

    dmd_name_string = 'RGB DMD'

    def _get_validated_config(self, config: dict) -> dict:
        return self.mc.config_validator.validate_config('rgb_dmds', config)

    @staticmethod
    def _reorder_channels(data, order):
        new_data = bytearray()
        for r, g, b in struct.iter_unpack('BBB', data):
            for channel in order:
                if channel == "r":
                    new_data.append(r)
                elif channel == "g":
                    new_data.append(g)
                elif channel == "b":
                    new_data.append(b)
                else:
                    raise ValueError("Unknown channel {}".format(channel))

        return bytes(new_data)

    def send(self, data: bytes) -> None:
        """Send data to RGB DMD via BCP."""
        if self.config['channel_order'] != 'rgb':
            data = self._reorder_channels(data, self.config['channel_order'])
        self.mc.bcp_processor.send('rgb_dmd_frame', rawbytes=data, name=self.name)



"""Handles key strokes in the media manager."""
import logging

from kivy.core.window import Window
from kivy.uix.widget import Widget as KivyWidget


class Keyboard(KivyWidget):

    """Handles key strokes in the media manager."""

    def __init__(self, mc, **kwargs):
        super().__init__(**kwargs)

        self.mc = mc
        self.log = logging.getLogger("Keyboard")

        self.keyboard_events = list()
        self.key_map = dict()
        self.toggle_keys = set()
        self.inverted_keys = list()
        self.active_keys = dict()
        self.debug = False

        # todo need to change the parsing logic to make these work
        self.special_keys = {
            'equals': '=',
            'minus': '-',
            'dash': '-',
            'leftbracket': '[',
            'rightbracket': ']',
            'backslash': '\\',
            'apostrophe': "'",
            'semicolon': ';',
            'colon': ':',
            'comma': ',',
            'period': '.',
            'slash': '/',
            'question': '?',
        }

        self.keyboard = Window.request_keyboard(callback=None, target=self)

        self.keyboard.bind(on_key_down=self._on_keyboard_down,
                           on_key_up=self._on_keyboard_up)

        for k, v in self.mc.machine_config['keyboard'].items():

            if k == 'debug' and v:
                self.debug = True
                continue

            k = str(k)  # k is the value of the key entry in the config
            switch_name = v.get('switch', None)
            # set whether a key is the push on / push off type
            toggle_key = v.get('toggle', None)
            invert = v.get('invert', None)
            event = v.get('event', None)
            mc_event = v.get('mc_event', None)
            params = v.get('params', None)
            # todo add args processing?

            # Process the key map entry
            k = k.replace('+', '-').lower().split('-')
            key = k[-1]
            mods = k[:-1]

            if mods:
                mods = sorted(mods)

            # What happens when it's pressed?
            if switch_name:  # We're processing a key entry for a switch
                if invert:
                    self.inverted_keys.append(switch_name)
                self.add_key_map(key, mods, switch_name, toggle_key)

            elif event:  # we're processing an entry for an event
                event_dict = {'event': event, 'params': params}
                self.add_key_map(key, mods, event_dict=event_dict)

            elif mc_event:  # we're processing an entry for an mc_event
                event_dict = {'mc_event': mc_event, 'params': params}
                self.add_key_map(key, mods, event_dict=event_dict)

    @staticmethod
    def get_key_string(key: str, mods: [str]) -> str:
        """Return string for key + modifiers.

        Args:
            key: string of key
            mods: list of modifiers as string

        Returns: string
        """
        try:
            mods = sorted(mods)
        except AttributeError:
            pass

        return '{}-{}'.format(key, '-'.join(mods))

    # pylint: disable-msg=too-many-arguments
    def add_key_map(self, key, mods, switch_name=None, toggle_key=False,
                    event_dict=None):
        """Add an entry to the key_map which is used to see what to do when key events are received.

        Args:
            key: The character or name of the key
            mods: List of strings for modifier keys for this entry
            switch_name: String name of the switch this key combination is tied
                to.
            toggle_key: Boolean as to whether this key should be a toggle key.
                (i.e. push on / push off).
            event_dict: Dictionary of events with parameters that will be
            posted when this key combination is pressed. Default is None.
        """
        key_string = self.get_key_string(key, mods)

        if switch_name:
            self.key_map[key_string] = switch_name
        elif event_dict:
            self.key_map[key_string] = event_dict

        if toggle_key:
            self.toggle_keys.add(key_string)

    def _on_keyboard_up(self, keyboard, keycode):
        del keyboard
        key = keycode[1]
        self.process_key_release(key)
        return True

    def _on_keyboard_down(self, keyboard, keycode, text, modifiers):
        del keyboard
        del text
        key = keycode[1]

        if self.debug:
            if modifiers:
                print("KEYS: {}+{}".format('+'.join(modifiers), key))
            else:
                print("KEYS: {}".format(key))

        if key in self.active_keys:
            return True
        else:
            return self.process_key_down(key, modifiers)

    def process_key_down(self, key, mods):
        """Process a key down event and change switches accordingly."""
        key_string = self.get_key_string(key, mods)
        self.log.debug("Processing key stroke for key %s", key_string)

        if key_string not in self.key_map:
            return False

        if key_string in self.toggle_keys:  # is this is a toggle key?
            self.active_keys[key] = None
            self.send_switch(state=-1, name=self.key_map[key_string])

        else:
            # do we have an event or a switch?
            if isinstance(self.key_map[key_string], str):  # switch

                if self.key_map[key_string] in self.inverted_keys:
                    self.send_switch(state=0, name=self.key_map[key_string])
                    self.active_keys[key] = ''.join(('-',
                                                     self.key_map[key_string]))

                else:
                    self.send_switch(state=1, name=self.key_map[key_string])
                    self.active_keys[key] = self.key_map[key_string]

            elif isinstance(self.key_map[key_string], dict):  # event
                event_dict = self.key_map[key_string]
                event_params = event_dict['params'] or {}

                if 'event' in event_dict:
                    self.mc.post_mc_native_event(str(event_dict['event']),
                                                 **event_params)

                elif 'mc_event' in event_dict:
                    self.log.warning("The keyboard setting 'mc_event:' will "
                                     "go away soon since now regular 'event:'"
                                     "entries are posted to the MC and MPF.")
                    # todo deprecate this
                    self.mc.events.post(event_dict['mc_event'],
                                        **event_params)

        return True

    def process_key_release(self, key):
        """Process a key up event and change switches accordingly."""
        action = self.active_keys.pop(key, None)

        if action:
            if action.startswith('-'):
                self.send_switch(state=1, name=action[1:])
            else:
                self.send_switch(state=0, name=action)

    def send_switch(self, name, state):
        """Notify mpf via BCP about a switch change."""
        if self.mc.bcp_client_connected:
            self.mc.bcp_processor.send('switch', name=name, state=state)



from mpfmc.core.audio.sdl2 cimport *
from mpfmc.core.audio.gstreamer cimport *
from mpfmc.core.audio.sound_file cimport *
from mpfmc.core.audio.track cimport *
from mpfmc.core.audio.notification_message cimport *


# ---------------------------------------------------------------------------
#    Standard Track types
# ---------------------------------------------------------------------------

cdef enum:
    no_marker = 0xFFFFFFFF

ctypedef struct TrackStandardState:
    # State variables for TrackStandard tracks
    int sound_player_count
    SoundPlayer *sound_players

cdef enum SoundPlayerStatus:
    # Enumeration of the possible sound player status values.
    player_idle = 0
    player_pending = 1
    player_replacing = 2
    player_fading_in = 3
    player_fading_out = 4
    player_playing = 5
    player_finished = 6
    player_stopping = 7

ctypedef struct DuckingSettings:
    int track_bit_mask
    Sint32 attack_start_pos
    Sint32 attack_duration
    Uint8 attenuation_volume
    Sint32 release_start_pos
    Sint32 release_duration

cdef enum DuckingStage:
    ducking_stage_idle = 0
    ducking_stage_delay = 1
    ducking_stage_attack = 2
    ducking_stage_hold = 3
    ducking_stage_release = 4
    ducking_stage_finished = 5

cdef enum FadingStatus:
    fading_status_not_fading = 0
    fading_status_fading_in = 1
    fading_status_fading_out = 2

ctypedef struct SoundSettings:
    SoundSample *sample
    Uint8 volume
    Uint8 volume_left
    Uint8 volume_right
    int loops_remaining
    int current_loop
    Uint32 sample_pos
    Uint32 loop_start_pos
    Uint32 loop_end_pos
    Uint64 sound_id
    Uint64 sound_instance_id
    int sound_priority
    FadingStatus fading_status
    Uint32 fade_in_steps
    Uint32 fade_out_steps
    Uint32 fade_steps_remaining
    Uint8 marker_count
    GArray *markers
    Uint32 about_to_finish_marker
    bint sound_has_ducking
    DuckingSettings ducking_settings
    DuckingStage ducking_stage
    GArray *ducking_control_points

ctypedef struct SoundPlayer:
    # The SoundPlayer keeps track of the current sample position in the source audio
    # samples and is also keeps track of variables for sound looping and determining when the
    # sound has finished playing.
    SoundPlayerStatus status
    SoundSettings current
    SoundSettings next
    int track_num
    int number


# ---------------------------------------------------------------------------
#    TrackStandard class
# ---------------------------------------------------------------------------
cdef class TrackStandard(Track):

    cdef list _sound_queue
    cdef dict _playing_instances_by_id
    cdef int _max_simultaneous_sounds

    # Track state needs to be stored in a C struct in order for them to be accessible in
    # the SDL callback functions without the GIL (for performance reasons).
    # The TrackStandardState struct is allocated during construction and freed during
    # destruction.
    cdef TrackStandardState *type_state

    cdef int _get_playing_sound_count(self, Uint64 sound_id)
    cdef list _get_playing_sound_instances(self, Uint64 sound_id)
    cdef int _get_idle_sound_player(self)
    cdef process_notification_message(self, NotificationMessageContainer *notification_message)
    cdef tuple _get_sound_player_with_lowest_priority(self)
    cdef bint _play_sound_on_sound_player(self, sound_instance, int player, bint force=?)
    cdef _set_player_sound_settings(self, SoundSettings *sound_settings, object sound_instance)
    cdef _set_player_playing(self, SoundPlayer *player, object sound_instance)
    cdef _set_player_replacing(self, SoundPlayer *player, object sound_instance)
    cdef int _get_player_playing_sound_instance(self, sound_instance)
    cdef Uint32 _fix_sample_frame_pos(self, Uint32 sample_pos, Uint8 bytes_per_sample, int channels)

    @staticmethod
    cdef void mix_playing_sounds(TrackState *track, Uint32 buffer_length, AudioCallbackData *callback_data) nogil

# ---------------------------------------------------------------------------
#    Global C functions designed to be called from the static audio callback
#    function (these functions do not use the GIL).
# ---------------------------------------------------------------------------

cdef bint get_memory_sound_samples(SoundSettings *sound, Uint32 length, Uint8 *output_buffer, int channels,
                                   Uint8 volume, TrackState *track, int player_num) nogil
cdef bint get_streaming_sound_samples(SoundSettings *sound, Uint32 length, Uint8 *output_buffer, int channels,
                                      Uint8 volume, TrackState *track, int player_num) nogil

cdef inline void end_of_sound_processing(SoundPlayer* player,
                                         TrackState *track) nogil:
    """
    Determines the action to take at the end of the sound (loop or stop) based on
    the current settings.  This function should be called when a sound processing
    loop has reached the end of the source buffer.
    Args:
        player: SoundPlayer pointer
        track: TrackState pointer for the current track
    """
    # Check if we are at the end of the source sample buffer (loop if applicable)
    if player.current.loops_remaining > 0:
        # At the end and still loops remaining, loop back to the beginning
        player.current.loops_remaining -= 1
        player.current.sample_pos = 0
        player.current.current_loop += 1
        send_sound_looping_notification(player.number,
                                 player.current.sound_id, player.current.sound_instance_id,
                                 track)

    elif player.current.loops_remaining == 0:
        # At the end and not looping, the sample has finished playing
        player.status = player_finished

    else:
        # Looping infinitely, loop back to the beginning
        player.current.sample_pos = 0
        player.current.current_loop += 1
        send_sound_looping_notification(player.number,
                                 player.current.sound_id, player.current.sound_instance_id,
                                 track)



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestAnimatedImages(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/animated_images'

    def get_config_file(self):
        return 'test_animated_images.yaml'

    def test_animated_images_loading(self):

        self.assertEqual(self.mc.images['ball'].image._anim_index, 0)
        self.assertEqual(self.mc.images['ball'].image._anim_delay, -1)
        self.assertEqual(self.mc.images['busy-stick-figures-animated'].
                         image._anim_index, 0)
        self.assertEqual(self.mc.images['busy-stick-figures-animated'].
                         image._anim_delay, -1)
        self.assertEqual(len(self.mc.images['busy-stick-figures-animated'].
                             image.image.textures), 10)

    def test_animated_images(self):
        self.mc.events.post('slide1')
        self.advance_time()

        ball = self.mc.targets['default'].current_slide.widgets[0].widget
        stick_figures = self.mc.targets['default'].current_slide.widgets[1].widget

        self.advance_time()

        # make sure they're playing as they should
        self.assertEqual(ball.fps, 30)
        self.assertEqual(ball.loops, -1)
        self.assertEqual(stick_figures.fps, 10)
        self.assertEqual(stick_figures.loops, -1)

        # test stopping
        stick_figures.stop()
        self.advance_time()
        stopped_frame = stick_figures.current_frame

        for x in range(10):
            self.assertEqual(stick_figures.current_frame, stopped_frame)
            self.advance_time()

        # test jumping to a new frame
        stick_figures.current_frame = 5
        self.assertEqual(stick_figures.current_frame, 6)

        # test starting
        stick_figures.play()
        self.advance_time()
        self.assertTrue(self.mc.images["busy-stick-figures-animated"].image._anim_ev)
        #self.assertEqual(1, self.mc.images["busy-stick-figures-animated"].references)

        self.mc.events.post('slide2')
        self.advance_time(.2)
        self.assertTrue(self.mc.images["busy-stick-figures-animated"].image._anim_ev)
        #self.assertEqual(2, self.mc.images["busy-stick-figures-animated"].references)

        self.mc.events.post('slide1_remove')
        self.advance_time(1)
        self.assertTrue(self.mc.images["busy-stick-figures-animated"].image._anim_ev)
        #self.assertEqual(1, self.mc.images["busy-stick-figures-animated"].references)



"""Copy of kivy.uix.effectwidget.

Remove when https://github.com/kivy/kivy/pull/5679 is merged and released.
"""
from kivy.graphics.context_instructions import PushMatrix, Color, PopMatrix
from kivy.graphics.fbo import Fbo
from kivy.graphics.gl_instructions import ClearBuffers, ClearColor
from kivy.graphics.instructions import RenderContext, Callback
from kivy.graphics.vertex_instructions import Rectangle
from kivy.uix.effectwidget import EffectFbo

from kivy.base import EventLoop
from kivy.properties import ListProperty, ObjectProperty

from kivy.uix.relativelayout import RelativeLayout


class EffectWidget(RelativeLayout):
    '''
    Widget with the ability to apply a series of graphical effects to
    its children. See the module documentation for more information on
    setting effects and creating your own.
    '''

    background_color = ListProperty((0, 0, 0, 0))
    '''This defines the background color to be used for the fbo in the
    EffectWidget.

    :attr:`background_color` is a :class:`ListProperty` defaults to
    (0, 0, 0, 0)
    '''

    texture = ObjectProperty(None)
    '''The output texture of the final :class:`~kivy.graphics.Fbo` after
    all effects have been applied.

    texture is an :class:`~kivy.properties.ObjectProperty` and defaults
    to None.
    '''

    effects = ListProperty([])
    '''List of all the effects to be applied. These should all be
    instances or subclasses of :class:`EffectBase`.

    effects is a :class:`ListProperty` and defaults to [].
    '''

    fbo_list = ListProperty([])
    '''(internal) List of all the fbos that are being used to apply
    the effects.

    fbo_list is a :class:`ListProperty` and defaults to [].
    '''

    _bound_effects = ListProperty([])
    '''(internal) List of effect classes that have been given an fbo to
    manage. This is necessary so that the fbo can be removed if the
    effect is no longer in use.

    _bound_effects is a :class:`ListProperty` and defaults to [].
    '''

    def __init__(self, **kwargs):
        # Make sure opengl context exists
        EventLoop.ensure_window()

        self.canvas = RenderContext(use_parent_projection=True,
                                    use_parent_modelview=True)
        self._callbacks = {}

        with self.canvas:
            self.fbo = Fbo(size=self.size)

        with self.fbo.before:   # noqa
            PushMatrix()
        with self.fbo:
            ClearColor(0, 0, 0, 0)
            ClearBuffers()
            self._background_color = Color(*self.background_color)
            self.fbo_rectangle = Rectangle(size=self.size)
        with self.fbo.after:    # noqa
            PopMatrix()

        super(EffectWidget, self).__init__(**kwargs)

        fbind = self.fbind
        fbo_setup = self.refresh_fbo_setup
        fbind('size', fbo_setup)
        fbind('effects', fbo_setup)
        fbind('background_color', self._refresh_background_color)

        self.refresh_fbo_setup()
        self._refresh_background_color()  # In case thi was changed in kwargs

    def _refresh_background_color(self, *args):     # noqa
        self._background_color.rgba = self.background_color

    def _propagate_updates(self, *largs):
        """Propagate updates from widgets."""
        del largs
        for fbo in self.fbo_list:
            fbo.ask_update()

    def refresh_fbo_setup(self, *args):     # noqa
        '''(internal) Creates and assigns one :class:`~kivy.graphics.Fbo`
        per effect, and makes sure all sizes etc. are correct and
        consistent.
        '''
        # Add/remove fbos until there is one per effect
        while len(self.fbo_list) < len(self.effects):
            with self.canvas:
                new_fbo = EffectFbo(size=self.size)
            with new_fbo:
                ClearColor(0, 0, 0, 0)
                ClearBuffers()
                Color(1, 1, 1, 1)
                new_fbo.texture_rectangle = Rectangle(size=self.size)

                new_fbo.texture_rectangle.size = self.size
            self.fbo_list.append(new_fbo)
        while len(self.fbo_list) > len(self.effects):
            old_fbo = self.fbo_list.pop()
            self.canvas.remove(old_fbo)

        # Remove fbos from unused effects
        for effect in self._bound_effects:
            if effect not in self.effects:
                effect.fbo = None
        self._bound_effects = self.effects

        # Do resizing etc.
        self.fbo.size = self.size
        self.fbo_rectangle.size = self.size
        for i in range(len(self.fbo_list)):
            self.fbo_list[i].size = self.size
            self.fbo_list[i].texture_rectangle.size = self.size

        # If there are no effects, just draw our main fbo
        if len(self.fbo_list) == 0:     # noqa
            self.texture = self.fbo.texture
            return

        for i in range(1, len(self.fbo_list)):
            fbo = self.fbo_list[i]
            fbo.texture_rectangle.texture = self.fbo_list[i - 1].texture

        # Build effect shaders
        for effect, fbo in zip(self.effects, self.fbo_list):
            effect.fbo = fbo

        self.fbo_list[0].texture_rectangle.texture = self.fbo.texture
        self.texture = self.fbo_list[-1].texture

        for fbo in self.fbo_list:
            fbo.draw()
        self.fbo.draw()

    def add_widget(self, widget):   # noqa
        # Add the widget to our Fbo instead of the normal canvas
        c = self.canvas
        self.canvas = self.fbo
        with widget.canvas:
            self._callbacks[widget.canvas] = Callback(self._propagate_updates)
        super(EffectWidget, self).add_widget(widget)
        self.canvas = c

    def remove_widget(self, widget):
        # Remove the widget from our Fbo instead of the normal canvas
        c = self.canvas
        widget.canvas.remove(self._callbacks[widget.canvas])
        del self._callbacks[widget.canvas]
        self.canvas = self.fbo
        super(EffectWidget, self).remove_widget(widget)
        self.canvas = c

    def clear_widgets(self, children=None):
        # Clear widgets from our Fbo instead of the normal canvas
        c = self.canvas
        self.canvas = self.fbo
        for canvas, callback in self._callbacks.items():
            canvas.remove(callback)
        self._callbacks = {}

        super(EffectWidget, self).clear_widgets(children)
        self.canvas = c



from mpfmc.core.audio.sdl2 cimport *
from mpfmc.core.audio.gstreamer cimport *


# ---------------------------------------------------------------------------
#    Track-related types
# ---------------------------------------------------------------------------

# The number of control points per audio buffer (sets control rate for ducking)
cdef enum:
    CONTROL_POINTS_PER_BUFFER = 8

cdef enum:
    MAX_SIMULTANEOUS_SOUNDS_DEFAULT = 8

cdef enum:
    MAX_SIMULTANEOUS_SOUNDS_LIMIT = 32


cdef enum TrackStatus:
    track_status_stopped = 0
    track_status_stopping = 1
    track_status_playing = 2
    track_status_pausing = 3
    track_status_paused = 4

ctypedef struct TrackState:
    # Common track state variables (for all track types)
    AudioCallbackData *callback_data
    void (*mix_callback_function)(TrackState *track, Uint32 buffer_length, AudioCallbackData *callback_data) nogil
    void* type_state
    TrackStatus status
    bint active
    int number
    Uint8 volume
    Uint8 fade_volume_start
    Uint8 fade_volume_target
    Uint8 fade_volume_current
    Uint32 fade_steps
    Uint32 fade_steps_remaining
    int buffer_size
    Uint8 *buffer
    GSList *notification_messages
    bint ducking_is_active
    GArray* ducking_control_points


# ---------------------------------------------------------------------------
#    Track base class
# ---------------------------------------------------------------------------
cdef class Track:

    cdef str _name
    cdef int _number
    cdef list _events_when_stopped
    cdef list _events_when_played
    cdef list _events_when_paused
    cdef object mc
    cdef SDL_AudioDeviceID device_id
    cdef object log

    # Track attributes need to be stored in a C struct in order for them to be accessible in
    # the SDL callback functions without the GIL (for performance reasons).  The TrackState
    # struct is allocated during construction and freed during destruction.
    cdef TrackState *state

    cdef TrackState *get_state(self)

    # ---------------------------------------------------------------------------
    #    Static C functions designed to be called from the static audio callback
    #    function (these functions do not use the GIL).
    # ---------------------------------------------------------------------------
    @staticmethod
    cdef void mix_track_to_output(TrackState *track, AudioCallbackData* callback_data,
                                  Uint8 *output_buffer, Uint32 buffer_length) nogil

    @staticmethod
    cdef void mix_audio(Uint8* output_buffer, const Uint8* input_buffer, Uint32 buffer_length, int volume) nogil

    @staticmethod
    cdef void mix_audio_stereo(Uint8* output_buffer, const Uint8* input_buffer, Uint32 buffer_length,
                               int volume_left, int volume_right) nogil

    @staticmethod
    cdef void apply_volume(Uint8* output_buffer, const Uint8* input_buffer, Uint32 buffer_length, int volume) nogil

    @staticmethod
    cdef void apply_volume_stereo(Uint8* output_buffer, const Uint8* input_buffer, Uint32 buffer_length,
                                  int volume_left, int volume_right) nogil



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestFonts(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/fonts'

    def get_config_file(self):
        return 'built_in_dmd_fonts.yaml'

    def test_dmd_font_styles(self):

        self.mc.events.post('window_slide')
        self.advance_time(1)

        self.mc.events.post('dmd_big')
        self.advance_time(1)

        # Make sure the dmd_big style hasn't changed
        self.assertEqual(self.mc.machine_config['widget_styles']['big']['adjust_bottom'], 5)
        self.assertEqual(self.mc.machine_config['widget_styles']['big']['adjust_top'], 2)
        self.assertEqual(self.mc.machine_config['widget_styles']['big']['font_size'], 10)

        # make sure the adjustments are correct
        top_widget = self.mc.displays['dmd'].current_slide.widgets[0].widget
        bottom_widget = self.mc.displays['dmd'].current_slide.widgets[1].widget

        self.assertEqual(bottom_widget.anchor_y, 'bottom')
        self.assertEqual(
            self.mc.machine_config['widget_styles']['big']['adjust_bottom'],
            -bottom_widget.anchor_offset_pos[1])

        self.assertEqual(top_widget.anchor_y, 'top')
        self.assertEqual(top_widget.anchor_offset_pos[1], -top_widget.height +
                         self.mc.machine_config['widget_styles']['big']['adjust_top'])

        self.mc.events.post('dmd_med')
        self.advance_time(1)

        # Make sure the dmd_med style hasn't changed
        self.assertEqual(self.mc.machine_config['widget_styles']['medium']['adjust_bottom'], 1)
        self.assertEqual(self.mc.machine_config['widget_styles']['medium']['adjust_top'], 1)
        self.assertEqual(self.mc.machine_config['widget_styles']['medium']['font_size'], 8)

        # make sure the adjustments are correct
        top_widget = self.mc.displays['dmd'].current_slide.widgets[0].widget
        bottom_widget = self.mc.displays['dmd'].current_slide.widgets[1].widget

        self.assertEqual(bottom_widget.anchor_y, 'bottom')
        self.assertEqual(
            self.mc.machine_config['widget_styles']['medium']['adjust_bottom'],
            -bottom_widget.anchor_offset_pos[1])

        self.assertEqual(top_widget.anchor_y, 'top')
        self.assertEqual(top_widget.anchor_offset_pos[1], -top_widget.height +
                         self.mc.machine_config['widget_styles']['medium']['adjust_top'])

        self.mc.events.post('dmd_small')
        self.advance_time(1)

        # Make sure the dmd_small style hasn't changed
        self.assertEqual(self.mc.machine_config['widget_styles']['small']['adjust_bottom'], 3)
        self.assertEqual(self.mc.machine_config['widget_styles']['small']['adjust_top'], 2)
        self.assertEqual(self.mc.machine_config['widget_styles']['small']['font_size'], 9)

        # make sure the adjustments are correct
        top_widget = self.mc.displays['dmd'].current_slide.widgets[0].widget
        bottom_widget = self.mc.displays['dmd'].current_slide.widgets[1].widget

        self.assertEqual(bottom_widget.anchor_y, 'bottom')
        self.assertEqual(
            self.mc.machine_config['widget_styles']['small']['adjust_bottom'],
            -bottom_widget.anchor_offset_pos[1])

        self.assertEqual(top_widget.anchor_y, 'top')
        self.assertEqual(top_widget.anchor_offset_pos[1], -top_widget.height +
                         self.mc.machine_config['widget_styles']['small']['adjust_top'])



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestDisplaySingle(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/display'

    def get_config_file(self):
        return 'test_display_none.yaml'

    def test_mc_display_none(self):
        # Make sure a single display is loaded properly:

        self.assertIn('default', self.mc.displays)
        self.assertEqual(self.mc.displays['default'].size, [800, 600])
        self.assertEqual(self.mc.targets['default'],
                         self.mc.displays['default'])




from typing import Union, List
from mpfmc.core.config_collection import ConfigCollection

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.uix.widget import Widget


class AnimationCollection(ConfigCollection):

    config_section = 'animations'
    collection = 'animations'
    class_label = 'Animations'

    def process_config(self, config: Union[dict, list]) -> List["Widget"]:
        # processes the 'animations' section of a config file to populate the
        # mc.animation_configs dict.

        # config is localized to 'animations' section

        if isinstance(config, dict):
            config = [config]

        # iterate and build our final processed list
        new_list = list()
        for animation in config:
            new_list.append(self.process_animation(animation, self.mc.config_validator))

        return new_list

    @staticmethod
    def process_animation(config: dict, config_validator) -> dict:
        # config is localized to a single animation's settings within a list

        # str means it's a named animation
        if isinstance(config, str):
            config = dict(named_animation=config)

        # dict is settings for an animation
        elif isinstance(config, dict):
            config_validator.validate_config('widgets:animations', config)

            if len(config['property']) > len(config['value']):
                raise ValueError('Animation must have at least the same number of settings '
                                 'in its "value" list ({}) as the number settings '
                                 'in its "property" list ({}).'.
                                 format(config['value'], config['property']))

        return config


collection_cls = AnimationCollection



from kivy.uix.effectwidget import EffectBase
from kivy.properties import NumericProperty, ListProperty


class DotFilterEffect(EffectBase):

    """GLSL effect to render an on-screen dot filter to look like individual round
    dots/pixels (simulating a DMD).

    This shader is based on code from a tutorial by Jason Gorski.
    http://www.lighthouse3d.com/opengl/ledshader/

    Potential bug: This GLSL filter seems to not work on Windows 10 running on
    VMware Fusion on Mac OS X. It works fine on Windows 10 native as well as
    Mac OS X Native.
    """

    width = NumericProperty(128)
    '''
    Sets the width in pixels of the display widget on the screen. Typically
    this is larger than the dots_x parameter.

    width is a :class:`~kivy.properties.NumericProperty` and
    defaults to 128.
    '''

    height = NumericProperty(32)
    '''
    Sets the height in pixels of the display widget on the screen. Typically
    this is larger than the dots_y parameter.

    height is a :class:`~kivy.properties.NumericProperty` and
    defaults to 32.
    '''

    dots_x = NumericProperty(128)
    '''
    Sets the number of dots in the horizontal (x) direction.

    dots_x is a :class:`~kivy.properties.NumericProperty` and
    defaults to 128.
    '''

    dots_y = NumericProperty(32)
    '''
    Sets the number of dots in the vertical (y) direction.

    dots_y is a :class:`~kivy.properties.NumericProperty` and
    defaults to 32.
    '''

    blur = NumericProperty(0.1)
    '''
    Sets the size of the blur around each pixel where it's blended with
    the background. The value is relative to the pixel. (e.g. a value of
    0.1 will add a 10% blur around the edge of each pixel.)

    blur is a :class:`~kivy.properties.NumericProperty` and
    defaults to 0.1.
    '''

    dot_size = NumericProperty(0.5)
    '''
    Sets the size of the circle for the dot/pixel relative to the size of the
    square bounding box of the dot. A size of 1.0 means that the diameter
    of the dot will be the same as its bounding box, in other words a
    size of 1.0 means that the dot will touch each other.

    dot_size is a :class:`~kivy.properties.NumericProperty` and
    defaults to 0.5.
    '''

    background_color = ListProperty([0.1, 0.1, 0.1, 1.0])
    '''
    A four-item tuple or list that represents the color of the space between the
    dots, in RGBA format with individual values as floats between 0.0 - 1.0. If
    you want the background to be transparent, set it to (0.0, 0.0, 0.0, 0.0).

    background_color is a :class:`~kivy.properties.ListProperty` and
    defaults to (0.1, 0.1, 0.1, 1.0) (which is 10% gray with 100% alpha/fully
    opaque).
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.do_glsl()

    def on_width(self, *args):
        self.do_glsl()

    def on_height(self, *args):
        self.do_glsl()

    def on_dots_x(self, *args):
        self.do_glsl()

    def on_dots_y(self, *args):
        self.do_glsl()

    def on_blur(self, *args):
        self.do_glsl()

    def on_dot_size(self, *args):
        self.do_glsl()

    def on_background_color(self, *args):
        self.do_glsl()

    def do_glsl(self):
        background_color = tuple(map(float, self.background_color))
        dot_size = min(self.width / self.dots_x, self.height / self.dots_y)

        self.glsl = dot_filter_glsl.format(float(self.blur),
                                           self.dot_size / 2.0,
                                           dot_size,
                                           float(self.width),
                                           float(self.height),
                                           *background_color)


dot_filter_glsl = '''
        float blur = {};
        float dotRadius = {};
        float dotSize = {};

        vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords)
        {{
            vec2 texCoordsStep = 1.0/(vec2({},{})/dotSize);
            vec2 dotRegionCoords = fract(tex_coords.xy/texCoordsStep);

            vec2 powers = pow(abs(dotRegionCoords - 0.5),vec2(2.0));
            float radiusSqrd = pow(dotRadius,2.0);
            float gradient = smoothstep(radiusSqrd-blur, radiusSqrd+blur, powers.x+powers.y);

            vec4 newColor = mix(color, vec4({}, {}, {}, {}), gradient);
            return newColor;
        }}
        '''

effect_cls = DotFilterEffect
name = 'dot_filter'



from mpfmc.tests.MpfIntegrationTestCase import MpfIntegrationTestCase
from mpfmc.tests.MpfSlideTestCase import MpfSlideTestCase
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestWidgetsAndSlides(MpfIntegrationTestCase, MpfFakeGameTestCase, MpfSlideTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'integration/machine_files/widgets_and_slides/'

    def test_animation_events(self):
        self.post_event("play_animation_show")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("Test Text")
        self.assertEqual(0.0, self.mc.targets["default"].current_slide.children[0].widget.rotate.angle)
        self.post_event("event_to_rotate_widget")
        self.advance_time_and_run(.1)
        self.assertEqual(90.0, self.mc.targets["default"].current_slide.children[0].widget.rotate.angle)

    def test_slide_with_vars(self):
        self.post_event("play_show_with_var")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("SLIDE WITH VAR Test123")

    def test_anonymous_slides(self):
        self.post_event("start_mode1")
        self.advance_time_and_run(.1)
        self.post_event("show_anonymous_slide1")
        self.advance_time_and_run(.1)

        # anonymous slide1 is on top (mode1)
        self.assertTextOnTopSlide("Anonymous Slide1")
        self.assertSlideOnTop("mode1.slide_player-show_anonymous_slide1")

        self.post_event("start_mode2")
        self.advance_time_and_run(.1)
        self.post_event("show_anonymous_slide2")
        self.advance_time_and_run(.1)

        # anonymous slide2 is on top (mode2)
        self.assertTextOnTopSlide("Anonymous Slide2")
        self.assertSlideOnTop("mode2.slide_player-show_anonymous_slide2")

        self.post_event("stop_mode2")
        self.advance_time_and_run(.1)

        # anonymous slide1 is on top again (mode1)
        self.assertTextOnTopSlide("Anonymous Slide1")
        self.assertSlideOnTop("mode1.slide_player-show_anonymous_slide1")

    def test_upper_case_mode(self):
        """This used to crash."""
        self.start_mode("Upper_Case_Mode")
        self.post_event("show_random_slide")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("random_slide")
        self.stop_mode("Upper_Case_Mode")
        self.advance_time_and_run(.1)

    def test_mode_start_from_mc(self):
        self.start_game()
        self.mc.post_mc_native_event("start_mode6")
        self.advance_time_and_run()
        self.assertModeRunning("mode6")
        self.assertTextOnTopSlide('Slide Mode 6')

    def test_condition_slide_player(self):
        self.start_game()
        self.mc.post_mc_native_event("start_mode7")
        self.advance_time_and_run()
        self.assertModeRunning("mode7")
        self.post_event_with_params("test_conditional", var=1)
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide('Condition 1')
        self.post_event("remove_conditional")
        self.post_event_with_params("test_conditional", var=2)
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide('Condition 2')

    def test_placeholders(self):
        self.post_event("play_slide_last_game_score")
        self.advance_time_and_run(.1)
        self.assertTextInSlide('Player1:  Player2: ', "slide_last_game_score")

        self.start_game()
        self.add_player()
        self.post_event("start_mode4")
        self.advance_time_and_run(.1)
        self.post_event("show_variable_slide")
        self.machine.variables.set_machine_var("test1", "asd")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("variable_slide")
        self.assertTextInSlide('MAIN TEXT 1:Test 2:7 3:1.75 4:asd', "variable_slide")
        self.machine.game.player["test_int"] = 42
        self.machine.game.player["test_float"] = 42.23
        self.machine.game.player["test_str"] = "1337"
        self.machine.game.player["score"] = 1337
        self.machine.variables.set_machine_var("test1", "l33t")
        self.advance_time_and_run(.1)
        self.assertTextInSlide('MAIN TEXT 1:1337 2:42 3:42.23 4:l33t', "variable_slide")
        self.drain_all_balls()
        self.advance_time_and_run(.1)
        self.machine.game.player["score"] = 42
        self.post_event("start_mode4")
        self.advance_time_and_run(.1)
        self.post_event("show_variable_slide")
        self.advance_time_and_run(.1)
        self.assertTextInSlide('MAIN TEXT 1:Test 2:7 3:1.75 4:l33t', "variable_slide")
        self.stop_game()
        self.advance_time_and_run(.1)

        self.post_event("play_slide_last_game_score")
        self.advance_time_and_run(.1)
        self.assertTextInSlide('Player1: 1337 Player2: 42', "slide_last_game_score")

        self.start_game()
        self.advance_time_and_run(.1)
        self.machine.game.player["score"] = 23
        self.advance_time_and_run(.1)
        self.stop_game()

        self.post_event("play_slide_last_game_score")
        self.advance_time_and_run(.1)
        self.assertTextInSlide('Player1: 23 Player2: ', "slide_last_game_score")

    def test_widget_on_slide_of_another_mode(self):
        self.post_event("start_mode1")
        self.advance_time_and_run()
        self.post_event("show_widget_mode1_on_slide_mode2")
        self.advance_time_and_run()

        self.assertSlideNotActive("slide_mode2")

        self.post_event("start_mode2")
        self.advance_time_and_run()
        self.post_event("show_slide_mode2")
        self.advance_time_and_run()

        self.assertSlideOnTop("slide_mode2")
        self.assertTextInSlide("Widget Mode 1", "slide_mode2")

    def test_slide_on_target_of_another_mode(self):
        self.post_event("start_mode1")
        self.advance_time_and_run()
        self.post_event("play_slide_mode1_on_frame_mode2")
        self.advance_time_and_run()

        self.assertTextNotOnTopSlide("Slide Mode 1")

        self.post_event("start_mode2")
        self.advance_time_and_run()
        self.post_event("show_slide_mode2_frame")
        self.advance_time_and_run()

        self.assertSlideOnTop("slide_mode2_frame")
        self.assertTextInSlide("Slide Mode 1", "slide_mode2_frame")
        self.assertTextOnTopSlide("Slide Mode 1")

        # remove slide frame
        self.post_event("remove_slide_mode2_frame")
        self.advance_time_and_run()

        self.assertTextNotOnTopSlide("Slide Mode 1")

        # add again
        self.post_event("show_slide_mode2_frame")
        self.advance_time_and_run()

        # text should be there again
        self.assertTextOnTopSlide("Slide Mode 1")

        self.post_event("remove_slide_mode1_on_frame_mode2")
        self.advance_time_and_run()

        # text should not be there
        self.assertTextNotOnTopSlide("Slide Mode 1")

        # add text
        self.post_event("play_slide_mode1_on_frame_mode2")
        self.advance_time_and_run()

        # text should be there again
        self.assertTextOnTopSlide("Slide Mode 1")

        # remove slide frame
        self.post_event("remove_slide_mode2_frame")
        self.advance_time_and_run()

        # add again
        self.post_event("show_slide_mode2_frame")
        self.advance_time_and_run()

        # text should be there again
        self.assertTextOnTopSlide("Slide Mode 1")

        # remove slide frame
        self.post_event("remove_slide_mode2_frame")
        self.advance_time_and_run()

        self.post_event("remove_slide_mode1_on_frame_mode2")
        self.advance_time_and_run()

        # add again
        self.post_event("show_slide_mode2_frame")
        self.advance_time_and_run()

        # text should not be there
        self.assertTextNotOnTopSlide("Slide Mode 1")

    def test_moving_slide_frame(self):
        self.post_event("start_mode3")
        self.advance_time_and_run()
        self.assertModeRunning("mode3")

        self.post_event("show_top_slide")
        self.advance_time_and_run()
        self.post_event("show_content_slide")
        self.advance_time_and_run()

        self.assertTextOnTopSlide("ASD")
        self.assertTextOnTopSlide("ASD2")

        self.assertAlmostEqual(16, self.mc.active_slides['top_slide'].widgets[0].widget.y)

        self.post_event("move_out")
        self.advance_time_and_run()
        self.assertAlmostEqual(-100.0, self.mc.active_slides['top_slide'].widgets[0].widget.y)
        self.post_event("move_back")
        self.advance_time_and_run()

        self.assertAlmostEqual(-134.0, self.mc.active_slides['top_slide'].widgets[0].widget.y)

    def test_dynamic_widget_width(self):
        self.start_game()
        self.start_mode("mode5")
        self.post_event("show_green_slide")
        self.advance_time_and_run(.1)
        self.assertEqual(0, self.mc.active_slides['green_slide'].widgets[0].widget.width)
        self.machine.game.player.greenness = 30
        self.advance_time_and_run(1.1)
        self.assertEqual(30, self.mc.active_slides['green_slide'].widgets[0].widget.width)
        self.machine.game.player.greenness = 100
        self.advance_time_and_run(1.1)
        self.assertEqual(100, self.mc.active_slides['green_slide'].widgets[0].widget.width)




import os
import sys

from mpf.core.case_insensitive_dict import CaseInsensitiveDict
from mpf.core.config_processor import ConfigProcessor
from mpf.core.utility_functions import Util


def percent_to_float(number_str, total):
    if str(number_str)[-1] == '%':
        return float(number_str[:-1]) * total / 100
    else:
        return float(number_str)


def center_of_points_list(points: list) -> tuple:
    """Calculates the center (average) of points in a list."""

    # Extract all x coordinates from list of points (odd list positions)
    coordinates_x = points[::2]

    # Extract all y coordinates from list of points (even list positions)
    coordinates_y = points[1::2]

    return sum(coordinates_x) / len(coordinates_x), sum(coordinates_y) / len(coordinates_y)
